"""From here:

https://github.com/huggingface/datasets/blob/a1b5a32365cd2c6631fa4ad236a881f9f62bf32b/src/datasets/commands/env.py
"""

import platform
from argparse import ArgumentParser

from trove import __version__ as version
from trove.commands import BaseTroveCLICommand


def info_command_factory(_):
    return EnvironmentCommand()


class EnvironmentCommand(BaseTroveCLICommand):
    @staticmethod
    def register_subcommand(parser: ArgumentParser):
        cmd_parser = parser.add_parser(
            "env", help="Print relevant system environment info."
        )
        cmd_parser.set_defaults(func=info_command_factory)

    def run(self):
        info = {
            "`trove` version": version,
            "Platform": platform.platform(),
            "Python version": platform.python_version(),
        }

        print("\nCopy-and-paste the text below in your GitHub issue.\n")
        print(self.format_dict(info))

        return info

    @staticmethod
    def format_dict(d):
        return "\n".join([f"- {prop}: {val}" for prop, val in d.items()]) + "\n"



import time
from datetime import timedelta

import torch.distributed as dist
from tqdm import tqdm

from ..logging_utils import get_logger_with_config

logger, logging_conf = get_logger_with_config("trove")


class InfiniteBarrier:
    def __init__(
        self,
        master_addr: str,
        port: int,
        world_size: int,
        rank: int,
        timeout: int = 600,
    ) -> None:
        """Alternative to ``dist.barrier()`` that does not timeout.

        Use it like this::

            inf_barrier = InfiniteBarrier(...) # craete the barrier
            inf_barrier.sync() # instead of dist.barrier()

        If for whatever reason, you are not able to change the timeout of the process group
        in distributed pytorch, you can use this class instead of ``dist.barrier()`` to avoid timeout errors.
        In general, you should write code that does not timeout in barrier but just in case!

        This class implements the barrier by a tcp key-value store. Each process that enters the barrier
        will increment the key of the active barrier by one. When the value of the key is equal to world size,
        the processes are released to continue.

        Args:
            master_addr: ip address of the master port.
            port: free port on master node to listen for messages from other processes.
            world_size: number of total processes involved.
            rank: global rank of current process.
            timeout: timeout in seconds for tcp store. You probably do not need to use this ever.
        """
        self.world_size = world_size
        self.process_index = rank
        # this increments by one for each new barrier block
        self.barrier_idx = -1
        if world_size > 1:
            self.kv_store = dist.TCPStore(
                master_addr, port, world_size, rank == 0, timedelta(seconds=timeout)
            )
        else:
            self.kv_store = None

    @property
    def _barrier_name(self) -> str:
        """Name of the current barrier in the key-value store."""
        return f"barrier_{self.barrier_idx}"

    def _enter_barrier(self) -> None:
        """Enter the key of the current barrier in the key-value store by 1 indicating that one
        more process has reached this barrier."""
        # you only enter this function once for each barrier in the code.
        # So if you enter this function, it is a new barrier with a new index and name
        self.barrier_idx += 1
        self.kv_store.add(self._barrier_name, 1)

    def _is_barrier_sync(self) -> bool:
        """Returns True if all processes have reached this barrier."""
        return int(self.kv_store.get(self._barrier_name)) == self.world_size

    def sync(self) -> None:
        """Blocks the process until all processes reach this part of the code."""
        if self.kv_store is None:
            # there is only process. No need for barrier
            return
        # after 20 minutes of waiting, print warning messages every 2 minutes.
        s = time.time()
        last_log = s

        # announce to other that one new process has entered the barrier
        self._enter_barrier()
        # wait for 0.2 seconds before checking again if all processes have reached this barrier
        t = 0.2
        # remain in loop until all processes enter this function
        while not self._is_barrier_sync():
            time.sleep(t)
            # next time sleep longer
            t *= 2
            if t > 20:
                # do not wait sleep more than 20 seconds
                t = 0.2
            c = time.time()
            if (c - s) >= 1200 and (c - last_log) >= 120:
                e_fmt = tqdm.format_interval(c - s)
                msg = f"Process have been waiting in barrier for 'elapsed_time: {e_fmt}'. Make sure it is expected."
                logger.warning(msg)
                last_log = c

        # make sure all processes exit the loop before returning.
        # If you return and the interpreter immediately exits but other processes has not
        # yet checked the kv store, that is gonna raise an exception
        # we just want to make sure all processes will be here soon (in probably less than a second)
        # and let torch handle the rest
        dist.barrier()



"""Configs for loading and processing the data.

From here:
https://github.com/texttron/tevatron/blob/7d298b428234f1c1065e98244827824753361815/src/tevatron/retriever/arguments.py
"""

from dataclasses import dataclass, fields
from typing import Dict, Optional


@dataclass
class DataArguments:

    dataset_name: Optional[str] = None
    """Name of the dataset. Only used if your query/passage formatting functions behave differently for different datasets"""
    group_size: int = 8
    """Number of passages used for each query during training or approximate evaluation during training
    (i.e., only used with RetrievalTrainer and NOT used with RetrievalEvaluator).
    """
    positive_passage_no_shuffle: bool = False
    """(for binary IR dataset) always use the first positive passage for training"""
    negative_passage_no_shuffle: bool = False
    """(for binary IR dataset) always use the first n negative passages for training"""
    passage_selection_strategy: str = "most_relevant"
    """(Only for MultiLevelDataset) How to choose a subset of passages for each query.
    Valid options are None, 'random', 'least_relevant', and 'most_relevant'.
    """
    query_max_len: Optional[int] = 32
    """The maximum total input sequence length after tokenization for query. Sequences longer
    than this will be truncated, sequences shorter will be padded.
    """
    passage_max_len: Optional[int] = 128
    """The maximum total input sequence length after tokenization for passage. Sequences longer
    than this will be truncated, sequences shorter will be padded.
    """
    pad_to_multiple_of: Optional[int] = 16
    """If set will pad the sequence to a multiple of the provided value. This is especially useful to
    enable the use of Tensor Cores on NVIDIA hardware with compute capability >= 7.5 (Volta).
    """

    def to_dict(self) -> Dict:
        """Return a json serializable view of the class attributes."""
        json_dict = dict()
        field_names = [f.name for f in fields(self)]
        for fname in field_names:
            fvalue = getattr(self, fname)
            json_dict[fname] = fvalue
        return json_dict



import json
from pathlib import Path
from typing import Dict, Optional

import huggingface_hub as hf_hub
import torch
import torch.distributed as dist
from peft import (
    LoraConfig,
    PeftModel,
    TaskType,
    get_peft_model,
    prepare_model_for_kbit_training,
)
from transformers import AutoModel, BitsAndBytesConfig, TrainingArguments

from .model_args import ModelArguments

# file that holds the trove retriever config and is saved next to weight checkpoints
TROVE_RETRIEVER_CONFIG_FILE = "trove_retriever_config.json"

# A subset of transformers.PretrainedModel attributes and methods to expose in wrappers
# to remain Compatible with a *subset* (and not all) of tools expecting a PreTrainedModel
DELEGATE_ATTRS_TO_WRAPPED = [
    "config",
    "is_loaded_in_4bit",
    "is_loaded_in_8bit",
    "gradient_checkpointing_enable",
    "gradient_checkpointing_disable",
    "enable_input_require_grads",
]


def add_model_apis_to_wrapper(wrapper, model) -> None:
    """Exposes some transformers model APIs through its wraper.

    This function exposes a subset of ``transformers.PretrainedModel`` attributes and methods in wrappers
    to remain Compatible with a **subset** (and not all) of tools expecting a PreTrainedModel like ``transformers.Trainer`` module.

    The added methods and attributes point to a method or attribute of the wrapped model with the same.

    Args:
        wrapper: the model that is wrapping a transformers model.
        model: an attribute of the ``wrapper`` class (often a transformers model).
    """
    global DELEGATE_ATTRS_TO_WRAPPED

    for attr in DELEGATE_ATTRS_TO_WRAPPED:
        if hasattr(model, attr):
            setattr(wrapper, attr, getattr(model, attr))


def load_trove_retriever_config(
    model_name: str, model_revision: Optional[str] = None
) -> Optional[Dict]:
    """Load the trove retriever config if present."""
    if Path(model_name).exists():
        # Model is saved locally
        config_path = Path(model_name).joinpath(TROVE_RETRIEVER_CONFIG_FILE)
        if not config_path.exists():
            return None
        assert config_path.exists()
    else:
        # Model is saved on hf hub
        assert hf_hub.repo_exists(repo_id=model_name)
        if not hf_hub.file_exists(
            repo_id=model_name,
            revision=model_revision,
            filename=TROVE_RETRIEVER_CONFIG_FILE,
        ):
            return None
        config_path = hf_hub.hf_hub_download(
            repo_id=model_name,
            revision=model_revision,
            filename=TROVE_RETRIEVER_CONFIG_FILE,
        )
    with open(config_path, "r") as f:
        config = json.load(f)
    return config


def find_base_name_or_path(model_name: str) -> Optional[str]:
    """Read the base model's name from model_name's config file."""
    if Path(model_name).exists():
        # Model is saved locally
        config_path = Path(model_name).joinpath("config.json")
        base_key = "_name_or_path"
        if not config_path.exists():
            # it is probably a peft model
            config_path = Path(model_name).joinpath("adapter_config.json")
            base_key = "base_model_name_or_path"
        assert config_path.exists()
    else:
        # Model is saved on hf hub
        assert hf_hub.repo_exists(repo_id=model_name)
        fname = "config.json"
        base_key = "_name_or_path"
        if not hf_hub.file_exists(repo_id=model_name, filename=fname):
            # it is probably a peft model
            fname = "adapter_config.json"
            base_key = "base_model_name_or_path"
        assert hf_hub.file_exists(repo_id=model_name, filename=fname)
        config_path = hf_hub.hf_hub_download(repo_id=model_name, filename=fname)

    with open(config_path, "r") as f:
        config = json.load(f)
    base_model_name = config.get(base_key, None)
    return base_model_name


def gather_tensors(
    tensor_obj: Optional[torch.Tensor], world_size: Optional[int], rank: Optional[int]
) -> torch.Tensor:
    """Gather tensors from all the processes in a distributed environment.

    see `this <https://github.com/texttron/tevatron/blob/7d298b428234f1c1065e98244827824753361815/src/tevatron/retriever/modeling/encoder.py#L99>`_

    Args:
        tensor_obj (Optional[torch.Tensor]): the tensor object to be gathered.
        world_size (Optional[int]): world size in the distributed environment
        rank (Optional[int]): rank of the process in the distributed environment

    Returns:
        The tensors across all processes concatenated.
    """
    if tensor_obj is None:
        return None

    tensor_obj = tensor_obj.contiguous()
    all_tensors = [torch.empty_like(tensor_obj) for _ in range(world_size)]
    dist.all_gather(all_tensors, tensor_obj)

    # Not sure, but I think this allows gradients to pass through:
    all_tensors[rank] = tensor_obj
    all_tensors = torch.cat(all_tensors, dim=0)
    return all_tensors


def load_adapter_config(
    model_name_or_path: str,
    revision: Optional[str] = None,
) -> Optional[Dict]:
    """Load peft adapter config from checkpoint path and returns None if adapter config file does
    not exist."""
    if Path(model_name_or_path).exists():
        # it is a path to a checkpoint
        adapter_path = Path(model_name_or_path) / "adapter_config.json"
        if not adapter_path.exists():
            return None
    else:
        # it is a model on hf hub.
        assert hf_hub.repo_exists(repo_id=model_name_or_path)
        if not hf_hub.file_exists(
            repo_id=model_name_or_path,
            filename="adapter_config.json",
            revision=revision,
        ):
            return None
        adapter_path = hf_hub.hf_hub_download(
            repo_id=model_name_or_path,
            revision=revision,
            filename="adapter_config.json",
        )
    with open(adapter_path, "r") as f:
        adapter_config = json.load(f)
    return adapter_config


def load_transformers_model(
    model_args: ModelArguments,
    training_args: Optional[TrainingArguments] = None,
    **kwargs,
):
    """Loads a huggingface transformers model.

    If the given checkpoints are only adapter weights, the adapters are merged into the main model.
    If specified by ``model_args`` attributes, this function quantizes the model and adds LORA adapters.

    Args:
        model_args: specifies how to load the model.
        training_args: It is used to setup gradient checkpointing arguments.
        **kwargs: passed to ``transformers.AutoModel.from_pretrained``.

    Returns:
        An instance of huggingface transformers model. Optionally with adapters.
    """
    if model_args.load_in_4bit or model_args.load_in_8bit:
        quant_conf = BitsAndBytesConfig(
            load_in_4bit=model_args.load_in_4bit,
            load_in_8bit=model_args.load_in_8bit,
            bnb_4bit_compute_dtype=model_args.torch_dtype,
            bnb_4bit_quant_type=model_args.bnb_4bit_quant_type,
            bnb_4bit_use_double_quant=model_args.use_bnb_nested_quant,
            bnb_4bit_quant_storage=model_args.torch_dtype,
        )
    else:
        quant_conf = None

    adapter_config = load_adapter_config(
        model_name_or_path=model_args.model_name_or_path,
        revision=model_args.model_revision,
    )

    if adapter_config is None:
        base_name_or_path = model_args.model_name_or_path
        base_rev = model_args.model_revision
        adapter_name_or_path = None
        adapter_rev = None
    else:
        if model_args.use_peft:
            raise NotImplementedError
        base_name_or_path = adapter_config["base_model_name_or_path"]
        base_rev = adapter_config["revision"]
        adapter_name_or_path = model_args.model_name_or_path
        adapter_rev = model_args.model_revision

    init_kwargs = dict(trust_remote_code=model_args.trust_remote_code)
    if model_args.attn_implementation is not None:
        init_kwargs["attn_implementation"] = model_args.attn_implementation
    if model_args.torch_dtype is not None:
        torch_dtype = model_args.torch_dtype
        if isinstance(torch_dtype, str) and torch_dtype != "auto":
            torch_dtype = getattr(torch, torch_dtype)
        init_kwargs["torch_dtype"] = torch_dtype

    if quant_conf is not None:
        init_kwargs["quantization_config"] = quant_conf

    model = AutoModel.from_pretrained(
        base_name_or_path, revision=base_rev, **init_kwargs, **kwargs
    )
    if adapter_name_or_path is not None:
        model = PeftModel.from_pretrained(
            model,
            adapter_name_or_path,
            revision=adapter_rev,
            trust_remote_code=model_args.trust_remote_code,
        )
        model = model.merge_and_unload()

    if model_args.use_peft and quant_conf is not None:
        prep_kwargs = dict()
        if training_args is not None:
            prep_kwargs["use_gradient_checkpointing"] = (
                training_args.gradient_checkpointing
            )
            prep_kwargs["gradient_checkpointing_kwargs"] = (
                training_args.gradient_checkpointing_kwargs
            )
        model = prepare_model_for_kbit_training(model, **prep_kwargs)
    elif model_args.use_peft and getattr(
        training_args, "gradient_checkpointing", False
    ):
        model.enable_input_require_grads()

    if model_args.use_peft:
        target_modules = model_args.lora_target_modules
        if not isinstance(target_modules, list):
            target_modules = target_modules.split(",")

        lora_config = LoraConfig(
            base_model_name_or_path=model_args.model_name_or_path,
            task_type=getattr(TaskType, model_args.lora_task_type),
            r=model_args.lora_r,
            lora_alpha=model_args.lora_alpha,
            lora_dropout=model_args.lora_dropout,
            target_modules=target_modules,
            inference_mode=False,
        )
        model = get_peft_model(model, lora_config)

    return model


# taken from here:
# https://github.com/texttron/tevatron/blob/7afa8067b958ca68689cb3e23792034f52ded656/src/tevatron/retriever/modeling/dense.py
def first_token_pooling(last_hidden_state: torch.Tensor, **kwargs) -> torch.Tensor:
    reps = last_hidden_state[:, 0]
    return reps


# taken from here:
# https://github.com/texttron/tevatron/blob/7afa8067b958ca68689cb3e23792034f52ded656/src/tevatron/retriever/modeling/dense.py
def last_token_pooling(
    last_hidden_state: torch.Tensor, attention_mask: torch.Tensor
) -> torch.Tensor:
    left_padding = attention_mask[:, -1].sum() == attention_mask.shape[0]
    if left_padding:
        reps = last_hidden_state[:, -1]
    else:
        sequence_lengths = attention_mask.sum(dim=1) - 1
        batch_size = last_hidden_state.shape[0]
        reps = last_hidden_state[
            torch.arange(batch_size, device=last_hidden_state.device), sequence_lengths
        ]
    return reps


# taken from here:
# https://github.com/texttron/tevatron/blob/7afa8067b958ca68689cb3e23792034f52ded656/src/tevatron/retriever/modeling/dense.py
def mean_pooling(
    last_hidden_state: torch.Tensor, attention_mask: torch.Tensor
) -> torch.Tensor:
    masked_hiddens = last_hidden_state.masked_fill(
        ~attention_mask[..., None].bool(), 0.0
    )
    reps = masked_hiddens.sum(dim=1) / attention_mask.sum(dim=1)[..., None]
    return reps



import itertools
import os
from contextlib import ExitStack, contextmanager
from pathlib import Path
from typing import Any, List, Optional, Union

import datasets
from datasets.fingerprint import Hasher

from .. import cache_manager
from ..logging_utils import get_logger_with_config
from .key_to_row_indices import KeyToRowIndices

logger, logging_conf = get_logger_with_config("trove")


class RowsByKeySingleSource:
    def __init__(
        self,
        dataset: Optional[datasets.Dataset] = None,
        filepath: Optional[os.PathLike] = None,
        key_field: Optional[str] = None,
        new_key_field: Optional[str] = None,
        add_filepath_field: bool = False,
        allow_duplicates: bool = True,
        unique_field_subset: Optional[Union[str, List[str]]] = None,
        ignore_main_data: bool = False,
        num_proc: Optional[int] = None,
    ) -> None:
        """A custom collection that retrieves a list of rows that have a specific value in the
        specified column.

        The data source can be a JSONL file or an HF ``datasets.Dataset`` instance.

        Args:
            dataset (Optional[datasets.Dataset]): Use this ``datasets.Dataset`` instance as source
            filepath (Optional[os.PathLike]): Read data from file
            key_field (Optional[str]): name of the field to use for querying rows (e.g., ``'qid'``)
            new_key_field (Optional[str]): If given, change the name of the key field to ``new_key_field``
            add_filepath_field (bool): If true, add a new field to each row that contains the name of the source file
            allow_duplicates (bool): Not implemented
            unique_field_subset (Optional[Union[str, List[str]]]): Not implemented
            ignore_main_data (bool): if true, do not return the actual data fields for each row and just
                return the metadata fields like value of ``'key_field'`` and ``'filepath'`` if it exists.
            num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.
        """
        # You should specify one and only one source of data
        if dataset is not None and filepath is not None:
            msg = f"Only one source of data for rows is supported. Got: 'dataset': '{dataset}' and 'filepath': '{filepath}'"
            raise ValueError(msg)

        if not allow_duplicates or unique_field_subset is not None:
            msg = "Checking for duplicates is not supported yet"
            raise NotImplementedError(msg)

        self.filepath = filepath
        self.add_filepath_field = add_filepath_field
        self.ignore_main_data = ignore_main_data
        # A unique fingerprint for the contents and output of this collection
        # do not use this directly. Instead use 'self.fingerprint' attribute
        self._unique_fingerprint = None

        if dataset is not None:
            self.dataset = dataset
        elif self.filepath is not None:
            if Path(self.filepath).suffix not in [".json", ".jsonl"]:
                msg = "Only json/jsonl files are supported"
                raise NotImplementedError(msg)
            logger.debug(f"Reading file for RowsByKeySingleSource:\n{self.filepath}")
            self.dataset = datasets.load_dataset(
                "json", data_files=filepath, split="train"
            )
        else:
            # There are no records. It is an empty container
            k = "_id" if key_field is None else key_field
            self.dataset = datasets.Dataset.from_dict({k: []})

        if key_field is None:
            # If key_field is not explicitely provided, check if any of the predefined potential id fields are present
            ds_fields = self.dataset.column_names
            all_id_fields = [
                "_id",
                "qid",
                "query_id",
                "query-id",
                "docid",
                "passage_id",
                "passage-id",
                "corpus-id",
            ]
            ds_id_fields = list(set(ds_fields).intersection(set(all_id_fields)))
            if len(ds_id_fields) == 1:
                key_field = ds_id_fields[0]
            else:
                if len(ds_id_fields) == 0:
                    msg = "`key_field` is not explicitely specified and none of the predefined potential id fields are present"
                else:
                    msg = f"`key_field` is not explicitely specified and more than one predefined potential ID field found: {ds_id_fields}"
                raise RuntimeError(msg)

        assert key_field is not None and key_field in self.dataset.column_names

        if new_key_field is not None and new_key_field != key_field:
            # Change the name of the key column
            assert new_key_field not in self.dataset.column_names
            self.dataset = self.dataset.rename_columns({key_field: new_key_field})
            key_field = new_key_field

        self.key_field = key_field

        # Create a mapping from values of column 'key_field' to the list of indices of rows that contain those values
        self.key_to_row_idx = KeyToRowIndices(
            dataset=self.dataset, id_field=self.key_field, num_proc=num_proc
        )

        # Add a new field named self.filepath_field that holds the name of the source file.
        self.filepath_field = "_file"
        if self.add_filepath_field and self.filepath_field in self.dataset.column_names:
            msg = (
                "You've asked to add the filename of the source data as a new key (`_file`) to each record."
                "But field `_file` already exists. We do not overwrite existing data."
            )
            raise RuntimeError(msg)

    @property
    def fingerprint(self) -> str:
        """Calculates a unique fingerprint for the contents and output of this container.

        Containers with the same fingerprint generate the same output.
        """
        if self._unique_fingerprint is not None:
            return self._unique_fingerprint
        # The combined hash of the following attributes uniquely identifies this container
        # A container with the same hash always produces the same output
        # Technically, 'self.ignore_main_data' should also be included in the hash since it impacts the output
        # But, because 'self.ignore_main_data' can change multiple times during a single execution, we ignore it for now.
        hasher = Hasher()
        hasher.update(self.filepath)
        hasher.update(self.add_filepath_field)
        hasher.update(cache_manager.hf_dataset_fingerprint(self.dataset)[0])
        hasher.update(self.key_field)
        hasher.update(self.filepath_field)

        self._unique_fingerprint = hasher.hexdigest()
        return self._unique_fingerprint

    def update_fingerprint(self) -> str:
        """Invalidate and recompute the unique fingerprint for this container."""
        self._unique_fingerprint = None
        return self.fingerprint

    def set_index_lookup_storage_type(self, storage: str) -> None:
        """Select if the key to row index lookup table should be stored in memory or in memory-
        mapped lmdb dict."""
        self.key_to_row_idx.set_storage_type(storage=storage)

    @contextmanager
    def meta_info_only(self):
        """A context manager to temporarily ignore the actual data and just return the metadata
        about rows."""
        orig_ignore_main_data = self.ignore_main_data
        self.ignore_main_data = True
        try:
            yield
        finally:
            self.ignore_main_data = orig_ignore_main_data

    def keys(self):
        """similar to ``dict.keys``"""
        return self.key_to_row_idx.keys()

    def __getitem__(self, key: Any) -> List[Any]:
        """Uses precomputed list of indices of all rows that their ``key_field == key`` to extract
        the subset of rows."""
        row_indices = self.key_to_row_idx[key]
        assert len(row_indices), "Key not found"
        if self.ignore_main_data:
            if self.add_filepath_field:
                rows = [
                    {self.key_field: key, self.filepath_field: self.filepath}
                    for _ in range(len(row_indices))
                ]
            else:
                rows = [{self.key_field: key} for _ in range(len(row_indices))]
            return rows
        rows = list()
        for idx in row_indices:
            rec = self.dataset[idx]
            # Double check that since we've created the key_to_idx mapping, the dataset has not changed.
            assert rec[self.key_field] == key
            if self.add_filepath_field:
                rec[self.filepath_field] = self.filepath
            rows.append(rec)
        return rows

    def get(self, key: Any, default: Optional[Any] = None) -> Optional[Any]:
        """similar to ``dict.get``"""
        if key in self.key_to_row_idx and len(self.key_to_row_idx[key]):
            return self[key]
        else:
            return default


class RowsByKey:
    def __init__(
        self,
        record_mapping: Optional[
            Union[RowsByKeySingleSource, List[RowsByKeySingleSource]]
        ] = None,
        dataset: Optional[Union[datasets.Dataset, List[datasets.Dataset]]] = None,
        filepath: Optional[Union[os.PathLike, List[os.PathLike]]] = None,
        key_field: Optional[str] = None,
        new_key_field: Optional[str] = None,
        add_filepath_field: bool = False,
        allow_duplicates_per_id_lazy: bool = True,
        allow_duplicates: bool = True,
        unique_field_subset: Optional[Union[str, List[str]]] = None,
        ignore_main_data: bool = False,
        num_proc: Optional[int] = None,
    ) -> None:
        """It combines the results of multiple
        :class:`~trove.containers.rows_by_key.RowsByKeySingleSource` instances.

        The rest of the arguments are the same as :class:`~trove.containers.rows_by_key.RowsByKeySingleSource`

        Args:
            record_mapping (Optional[Union[RowsByKeySingleSource, List[RowsByKeySingleSource]]]):
                a single or a list of ``RowsByKeySingleSource`` instances to use as data source
            dataset (Optional[Union[datasets.Dataset, List[datasets.Dataset]]]): a single or a list of datasets
            filepath (Optional[Union[os.PathLike, List[os.PathLike]]]): a single or a list of files as data source
            allow_duplicates_per_id_lazy (bool): If False, when calling ``__getitem__``, it raises an exception if there are more than one row for the key
        """
        self.allow_duplicates_per_id_lazy = allow_duplicates_per_id_lazy
        # A unique fingerprint for the contents and output of this collection
        # do not use this directly. Instead use 'self.fingerprint' attribute
        self._unique_fingerprint = None

        if (
            sum([record_mapping is not None, dataset is not None, filepath is not None])
            > 1
        ):
            msg = "You can at most use one of the 'record_mapping', 'dataset', or 'filepath' as data source."
            raise ValueError(msg)

        if record_mapping is not None:
            if not isinstance(record_mapping, list):
                record_mapping = [record_mapping]
            self.record_mappings = record_mapping
        elif dataset is not None or filepath is not None:
            # Creates lists even if there is only one item
            if dataset is not None:
                if not isinstance(dataset, list):
                    dataset = [dataset]
                filepath = [filepath] * len(dataset)
            elif filepath is not None:
                if not isinstance(filepath, list):
                    filepath = [filepath]
                dataset = [dataset] * len(filepath)

            # Create Mappings for each data source
            self.record_mappings = [
                RowsByKeySingleSource(
                    dataset=ds,
                    filepath=path,
                    key_field=key_field,
                    new_key_field=new_key_field,
                    add_filepath_field=add_filepath_field,
                    allow_duplicates=True,
                    unique_field_subset=None,
                    ignore_main_data=ignore_main_data,
                    num_proc=num_proc,
                )
                for ds, path in zip(dataset, filepath)
            ]
        else:
            self.record_mappings = []

        self.check_duplicates(
            allow_duplicates=allow_duplicates, unique_field_subset=unique_field_subset
        )

    @property
    def fingerprint(self) -> str:
        """Calculates a unique fingerprint for the contents and output of this container.

        Containers with the same fingerprint generate the same output.
        """
        # The content and output of this container depends on the content of its record mappings
        # So we use the combined fingerprint of its record mappings as its fingerprint
        if self._unique_fingerprint is not None:
            return self._unique_fingerprint
        hasher = Hasher()
        for rmap in self.record_mappings:
            hasher.update(rmap.fingerprint)

        self._unique_fingerprint = hasher.hexdigest()
        return self._unique_fingerprint

    def update_fingerprint(self) -> str:
        """Invalidate and recompute the unique fingerprint for this container."""
        self._unique_fingerprint = None
        return self.fingerprint

    def check_duplicates(
        self,
        allow_duplicates: bool = True,
        unique_field_subset: Optional[Union[str, List[str]]] = None,
    ) -> None:
        """Check that there are no duplicate rows across all collections."""
        if not allow_duplicates or unique_field_subset is not None:
            raise NotImplementedError

    def set_index_lookup_storage_type(self, storage: str) -> None:
        """Select if the key to row index lookup table should be stored in memory or in memory-
        mapped lmdb dict."""
        for rmap in self.record_mappings:
            rmap.set_index_lookup_storage_type(storage=storage)

    @contextmanager
    def meta_info_only(self):
        """A context manager to temporarily ignore the actual data and just return the metadata
        about rows."""
        with ExitStack() as stack:
            for rmap in self.record_mappings:
                stack.enter_context(rmap.meta_info_only())
            yield

    def keys(self):
        """Merges the keys for all data sources."""
        return itertools.chain.from_iterable(
            [rmap.keys() for rmap in self.record_mappings]
        )

    def get(self, key: Any, default: Optional[Any] = None) -> Optional[Any]:
        """Merges the results returned by each of the data sources."""
        rows = list()
        for rec_map in self.record_mappings:
            rows.extend(rec_map.get(key, []))
        if len(rows):
            assert self.allow_duplicates_per_id_lazy or len(rows) == 1
            return rows
        else:
            return default

    def __getitem__(self, key: Any) -> List[Any]:
        """Same as ``get()`` method but to mimic python ``dict[key]`` interface."""
        rows = self.get(key, None)
        # The key does not exist
        assert rows is not None, "Key not found"
        return rows



"""Logic for managing the univeral artifact cache of the package."""

import datetime
import json
import os
from dataclasses import asdict
from pathlib import Path
from typing import Dict, Optional, Tuple, Union
from uuid import uuid4

import datasets
import huggingface_hub as hf_hub
from datasets.fingerprint import Hasher

from . import config, file_utils

# This will be set to 'True' if the process temporary cache pardir is registered to be
# removed before interpreter shutdown
_TEMP_PARDIR_WILL_BE_REMOVED = False


def _register_temp_pardir_for_removal() -> None:
    """Register temporary cache pardir to be removed before interpreter shutdown."""
    global _TEMP_PARDIR_WILL_BE_REMOVED
    if _TEMP_PARDIR_WILL_BE_REMOVED:
        return
    _TEMP_PARDIR_WILL_BE_REMOVED = True
    # Since the temp cache dir is exclusive to each process, we can safely remove
    # it before exit even if there are other processes that are still running (in a distributed environment).
    temp_pardir = Path(get_cache_pardir(uuid4().hex, "temp")).parent.as_posix()
    file_utils.register_for_removal(temp_pardir)


def local_file_fingerprint(path: os.PathLike, fingerprint_src: str) -> Tuple[str, Dict]:
    """Create unique fingerprint and metadata for local file.

    Fingerprint is the hash of the combination of:

        * resolved absolute ``path``
        * time of most recent content modification of ``path``.

    You can also choose to calculate the hash from the content of the file rather than its metadata.

    We also return a metadata dict that contains some information about the fingerprint.
    E.g., the path of the source file and when the hash was created.

    You can use this info to create a unique cashe directory for all the artifacts that are generated
    from ``path``.

    Args:
        path (os.PathLike): filepath to create metadata for.
        fingerprint_src (Optional[str]): how to calculate a unique hash for input_data. Options are

            * ``path``: hash the combination of resolved absolute path and the last modified timestamp.
            * ``content``: calculate a non-cryptographic hash from binary content of the file.

    Returns:
        unique fingerprint and metadata about `path`
    """
    _realpath = file_utils.realpath(path)
    if fingerprint_src == "path":
        hasher = Hasher()
        hasher.update(_realpath)
        hasher.update(Path(_realpath).stat().st_mtime)
        file_hash = hasher.hexdigest()
    elif fingerprint_src == "content":
        file_hash = file_utils.hash_file_bytes(path=path)
    else:
        msg = f"Supported values for 'fingerprint_src' are: 'path' and 'content'. Got: '{fingerprint_src}'"
        raise ValueError(msg)

    metadata_dict = {
        "source_path": Path(path).as_posix(),
        "source_realpath": _realpath,
        "timestamp": datetime.datetime.now().isoformat(),
    }
    if Path(path).exists():
        metadata_dict["last_modification_time"] = (
            Path(file_utils.realpath(path)).stat().st_mtime
        )
    return file_hash, metadata_dict


def hf_dataset_fingerprint(dataset: datasets.Dataset) -> Tuple[str, Dict]:
    """Creae unique fingerprint and metadata for ``datasets.Dataset`` object.

    Fingerprint is the hash of the combination of

        * original finger print of the dataset. This should be enough on its own but we do some extra work
        * parent directory that contains the arrow cache files that back the dataset.
          We include this because it has some information about split, source file, etc.

    We also return a metadata dict that contains info about the dataset that was used to compute the hash.

    Similar to :func:`local_file_fingerprint`, You can use this info to create a unique
    cash directory for all the artifacts that are generated from ``dataset``.

    Args:
        dataset (datasets.Dataset): dataset to create metadata for.

    Returns:
        unique fingerprint and metadata about dataset.
    """
    hasher = Hasher()
    hasher.update(dataset._fingerprint)
    if len(dataset.cache_files):
        _ds_cache_root = Path(dataset.cache_files[0]["filename"]).parent.as_posix()
        hasher.update(file_utils.realpath(_ds_cache_root))
    ds_hash = hasher.hexdigest()
    metadata_dict = asdict(dataset.info)
    return ds_hash, metadata_dict


def get_cache_pardir(
    artifact_content: str, artifact_type: str, cache_uuid: Optional[str] = None
) -> os.PathLike:
    """Create the parent directory holding cache files with specific content and type.

    The parent directory follows this format: ``CACHE_ROOT/ARTIFACT_TYPE/ARTIFACT_CONTENT``.
    ``ARTIFACT_TYPE`` and ``ARTIFACT_CONTENT`` are often the same as input arguments to this function with some exceptions.
    Exceptions are explained in comments in the code.

    Args:
        artifact_content (str): content of the artifacts (e.g., ``embedding``, ``grouped_qrel``, ``key_to_row_idx``, etc.).
            Basically asking what you want to cache.
        artifact_type (str): The type of artifacts. There are three possibilities

            * ``final``: these are the final results of the target operation
            * ``intermediate``: These are intermediate results and are just needed for a short time until the results of the next step are ready.
            * ``temp``: these are temporary cache files. Once this session ends, you won't be able to access them again.
              Also this is not synchronized across processes. You get a different cache pardir in each process.
            * ``user``: these are cache files saved and loaded by trove users and should not impact the internal functionality of the library.
        cache_uuid (Optional[str]): If provided, use ``cache_uuid`` whenever a random unique ID is needed.
            When your ``artifact_type='temp'`` or ``artifact_content='embedding'``, you can provide the same value for this argument
            from all processes to ensure they all have the same cache_pardir (e.g., across different processes)

    Returns:
        A cache directory that holds artifacts of specific type and content.
    """
    supported_art_types = ["final", "intermediate", "temp", "user"]

    parts = []
    if artifact_type == "user":
        parts.append("user")
    elif artifact_type == "final":
        parts.append("final")
    elif artifact_type == "intermediate":
        parts.append("intermediate_artifacts")
    elif artifact_type == "temp":
        parts.append("tmp")
        # For temporary results, create a random directory that changes everytime.
        # It is like deleting the directory. But, subsequent runs will not crash if we fail to clear the temporary files properly.
        if cache_uuid is not None:
            parts.append(cache_uuid)
        else:
            parts.append(config.PROCESS_UUID)
            # make sure the temporary cache pardir is removed before exit
            _register_temp_pardir_for_removal()
    elif artifact_type not in supported_art_types:
        msg = (
            f"`artifact_type` is not recognized."
            f" Supported artifact types are: {supported_art_types}. Got '{artifact_type}'"
        )
        raise ValueError(msg)

    parts.append(artifact_content)
    if artifact_content == "embedding":
        if cache_uuid is not None:
            parts.append(cache_uuid)
        else:
            parts.append(uuid4().hex)

    cache_pardir = hf_hub.cached_assets_path(
        library_name="trove",
        namespace=parts[0],
        subfolder=parts[1],
        assets_dir=config.TROVE_CACHE_DIR,
    )
    cache_pardir = cache_pardir.joinpath(*parts[2:])
    return cache_pardir.as_posix()


def get_cache_dir(
    input_data: Optional[Union[os.PathLike, datasets.Dataset]] = None,
    cache_pardir: Optional[os.PathLike] = None,
    artifact_content: Optional[str] = None,
    artifact_type: Optional[str] = None,
    cache_uuid: Optional[str] = None,
    write_metadata: bool = True,
    fingerprint_src: Optional[str] = "content",
    fingerprint: Optional[str] = None,
    metadata: Optional[Dict] = None,
) -> Path:
    """Creates a unique directory to save cache files generated for a specific input data.

    Cache files are organized in two levels. First is ``cache_pardir``, which organizes the cache files
    based on their type and content. See :func:`get_cache_pardir` for more info.
    Then, inside ``cache_pardir``, there is a unique directory for each source of input data.
    All cache files generated from a specific source of ``input_data`` should be saved in the corresponding cache directory.

    Args:
        input_data (Optional[Union[os.PathLike, datasets.Dataset]]): The data that was used to generate the cache.
            It can be either a local file or a ``datasets.Dataset`` instance. It is used to generate a unique fingerprint and some
            metadata about input data. You can provide the fingerprint and metadata dict directly instead of ``input_data``.
        cache_pardir (Optional[os.PathLike]): cache_pardir based on ``artifact_content`` and ``artifact_type``.
            If not provided, create it from ``artifact_content`` and ``artifact_type``.
        artifact_content (Optional[str]): see :func:`get_cache_pardir`
        artifact_type (Optional[str]): see :func:`get_cache_pardir`
        cache_uuid (Optional[str]): see :func:`get_cache_pardir`
        write_metadata (bool): If True, write a json file in ``cache_dir`` with some info about ``input_data``.
        fingerprint_src (Optional[str]): how to calculate a unique hash for input_data if it is a local path on disk.
            See :func:`local_file_fingerprint` for details.
        fingerprint (Optional[str]): a unique fingerprint for the data source that was used to generate the cache.
            if not provided, it is calculated from the value of ``input_data``.
        metadata (Optional[Dict]): a dictionary with some information about that helps identify the ``input_data`` that
            lead to this cache files (e.g. filepath, modification time, etc.). If not provided, it is calculated from ``input_data``.

    Returns:
        A unique cache directory to save artifacts generated by processing ``input_data``
    """
    if (input_data is None) == (fingerprint is None or metadata is None):
        msg = (
            "You should provide exactly one of the 'input_data' or '(fingerprint, metadata)' arguments."
            f" Got: 'input_data': '{input_data}', 'fingerprint': '{fingerprint}', 'metadata': '{metadata}'"
        )
        raise ValueError(msg)

    if input_data is None:
        if fingerprint is None or metadata is None:
            msg = (
                "You should provide both 'fingerprint' and 'metadata'."
                f" Got: 'fingerprint': '{fingerprint}', 'metadata': '{metadata}'"
            )
            raise ValueError(msg)

        _fingerprint = fingerprint
        _info = metadata
    else:
        if isinstance(input_data, datasets.Dataset):
            _fingerprint, _info = hf_dataset_fingerprint(input_data)
        else:
            _fingerprint, _info = local_file_fingerprint(
                path=input_data, fingerprint_src=fingerprint_src
            )

    if cache_pardir is None:
        if artifact_content is None or artifact_type is None:
            msg = "You should provide either `cache_pardir` or `artifact_content` and `artifact_type`."
            raise ValueError(msg)
        cache_pardir = get_cache_pardir(
            artifact_content=artifact_content,
            artifact_type=artifact_type,
            cache_uuid=cache_uuid,
        )

    cache_dir = Path(cache_pardir, _fingerprint)
    metadata_path = cache_dir / "source_metadata.json"

    try:
        cache_dir.mkdir(exist_ok=True, parents=True)
    except (FileExistsError, NotADirectoryError):
        raise ValueError(
            f"Corrupted cache folder: cannot create directory because of an existing file ({cache_dir})."
        )

    if write_metadata:
        with file_utils.easyfilelock(metadata_path):
            if not metadata_path.exists():
                with open(metadata_path, "w") as f:
                    json.dump(_info, f, indent=2)

    return cache_dir



import copy
import heapq
from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

import torch

from ..logging_utils import get_logger_with_config

logger, logging_conf = get_logger_with_config("trove")


class ResultHeapq:
    def __init__(
        self,
        topk: int = 100,
        special_docids: Optional[
            Dict[str, Union[List[str], Dict[str, Union[int, float]]]]
        ] = None,
    ) -> None:
        """A simple class that uses heapq to keep track of the topk largest scores for each query.

        You can also use an instance of this class as a callable in a pytorch evaluation loop.

        Args:
            topk (int): Number of items to hold in the heapq for each query.
            special_docids (Optional[Dict]): special set of documents for each query
                to collect their similarity scores regardless of their ranking.
                Do not use unless necessary. It degrades the performance.
                It can be in qrel format (i.e., ``special_docids[qid][docid]=some_score``). score is ignored here.
                It can also be a mapping from qid to a list of docids, i.e., ``special_docids[qid] = [docid1, docid2, ...]``
        """
        self.topk = topk
        self.topk_docs = None
        self.special_docs = None
        self.reset_state()

        self.special_docids = defaultdict(set)
        if special_docids is not None:
            for qid in special_docids.keys():
                _docids = special_docids[qid]
                if isinstance(_docids, dict):
                    # special_docids is in nested_qrel format
                    _docids = list(_docids.keys())
                self.special_docids[qid].update(_docids)
        self.special_docids = dict(self.special_docids)

    def reset_state(self) -> None:
        """Clear the data collected so far."""
        self.topk_docs = defaultdict(list)
        self.special_docs = dict(qids=[], docids=[], scores=[])

    def add_triplet(self, qid: str, docid: str, score: Union[int, float]) -> None:
        """Add (docid, score) to the heapq for qid.

        Args:
            qid (str): query id
            docid (str): document id
            score (Union[int, float]): similarity score between query and document.
        """
        if len(self.topk_docs[qid]) < self.topk:
            heapq.heappush(self.topk_docs[qid], (score, docid))
        else:
            heapq.heappushpop(self.topk_docs[qid], (score, docid))

    def add_qrel_nested_dict(
        self,
        qrel_dict: Union[
            Dict[str, Dict[str, Union[int, float]]],
            Dict[str, Dict[str, Dict[str, Union[int, float]]]],
        ],
    ) -> None:
        """Import results in nested qrel dict format.

        The input is similar to the output of :meth:`as_qrel_nested_dict` method.

        Args:
            qrel_dict (Dict): results in nested qrel dict format,
                where ``qrel_dict[qid][docid]`` is the similarity score between ``qid`` and ``docid``.
                If ``qrel_dict`` contains keys ``topk_docs`` or ``special_docs``, it assumes that each of the
                ``qrel_dict['topk_docs']`` and ``qrel_dict['special_docs']`` is a separate qrel object and is merged into the corresponding internal collection.
                If none of these keys is available, it assumes that ``qrel_dict`` is one qrel object and is merged into ``self.topk_docs``.
        """
        _topk_docs = qrel_dict.get("topk_docs", None)
        _special_docs = qrel_dict.get("special_docs", None)
        if _topk_docs is None and _special_docs is None:
            _topk_docs = qrel_dict

        if _topk_docs is not None:
            for qid, qdata in _topk_docs.items():
                for docid, docscore in qdata.items():
                    self.add_triplet(qid=qid, docid=docid, score=docscore)

        if _special_docs is not None:
            for qid, qdata in _special_docs.items():
                self.special_docs["qids"].extend([qid] * len(qdata))
                self.special_docs["docids"].extend(qdata.keys())
                self.special_docs["scores"].extend(qdata.values())

    def __call__(
        self, scores: torch.Tensor, qids: List[str], docids: List[str], **kwargs
    ) -> None:
        """Adds the calculated scores between a batch of queries and passages.

        Intended to be used as a callable in a pytorch evaluation loop.

        Args:
            scores (torch.Tensor): similarity scores between query and documents.
                2D tensor of shape ``[NUM_QUERIES, NUM_DOCS]``
            qids (List[str]): List of query ids in the batch of length NUM_QUERIES
            docids (List[str]): List of document ids in the batch of length NUM_DOCS
            **kwargs: No effect. Just for compatibility with other classes.
        """
        # Iterating over all scores takes a long time
        # Instead use torch to extract the topk largest scores and only add those to the heapq
        topk_scores, topk_indices = torch.topk(
            scores, k=min(self.topk, scores.shape[1]), dim=-1, largest=True, sorted=True
        )
        topk_scores = topk_scores.cpu().tolist()
        topk_indices = topk_indices.cpu().tolist()

        for q_id, q_scores, q_doc_indices in zip(qids, topk_scores, topk_indices):
            for score, d_idx in zip(q_scores, q_doc_indices):
                self.add_triplet(qid=q_id, docid=docids[d_idx], score=score)

        if len(self.special_docids):
            doc_id2idx = {_id: i for i, _id in enumerate(docids)}
            row_indices = list()
            col_indices = list()
            for qidx, qid in enumerate(qids):
                if qid not in self.special_docids:
                    continue
                _cell_docids = list(self.special_docids[qid].intersection(doc_id2idx))
                if not _cell_docids:
                    continue
                _col_indices = [doc_id2idx[_id] for _id in _cell_docids]
                col_indices.extend(_col_indices)
                row_indices.extend([qidx] * len(_cell_docids))
                self.special_docs["qids"].extend([qid] * len(_cell_docids))
                self.special_docs["docids"].extend(_cell_docids)

            if len(row_indices):
                cell_scores = scores[row_indices, col_indices].cpu().tolist()
                self.special_docs["scores"].extend(cell_scores)

    def as_qrel_nested_dict(
        self, collection: Optional[str] = None
    ) -> Union[Dict[str, Dict[str, float]], Dict[str, Dict[str, Dict[str, float]]]]:
        """Export the collected results as nested qrel dicts.

        Args:
            collection (Optional[Union[str, List[str]]]): The type of results to return.
                It can be one of the following:

                * ``None`` : Just return topk most similar documents
                * ``'topk_docs'`` : similar to None
                * ``'special_docs'`` : just return the collected results for special documents
                * ``'all'`` : return a dict with keys ``'special_docs'`` and ``'topk_docs'`` containing the corresponding results

        Returns:
            The collected similarities so far in qrel format (i.e., ``qrel[qid][docid]=sim(qid, docid)``).
            If asked for multiple collections of results, it returns a mapping from collection type to qrel.
        """
        logger.debug("Export collected scores in qrel nested dict format.")

        res = dict()

        if collection is None or collection in ["topk_docs", "all"]:
            qrel_mapping = defaultdict(dict)
            for qid, qdata in self.topk_docs.items():
                for docscore, docid in qdata:
                    qrel_mapping[qid][docid] = docscore
            res["topk_docs"] = dict(qrel_mapping)

        if collection in ["special_docs", "all"]:
            qrel_mapping = defaultdict(dict)
            for qid, docid, docscore in zip(
                self.special_docs["qids"],
                self.special_docs["docids"],
                self.special_docs["scores"],
            ):
                qrel_mapping[qid][docid] = docscore
            res["special_docs"] = dict(qrel_mapping)

        if len(res) == 1:
            return list(res.values())[0]
        else:
            return res

    def as_sorted_lists(self, reverse: bool = False) -> Dict[str, List[Tuple]]:
        """Export the topk collected so far as a sorted list for each query.

        Args:
            reverse (bool): Is used to sort the scores like ``sorted(..., reversed=reversed)``

        Returns:
            a mapping from ``qid`` to a list of tuples of ``(doc_score, doc_id)``. Each list is sorted based on the ``doc_score`` field.
        """
        output = dict()
        for qid, qdata in self.topk_docs.items():
            output[qid] = sorted(qdata, key=lambda x: x[0], reverse=reverse)
        return output

    def export_result_dump(
        self, reset_state: bool = False
    ) -> Dict[str, Dict[str, List[Tuple[str, float]]]]:
        """Export all the results collected so far in the same format as stored in this class.

        You can use this to export and merge the collected results from multiple collections,
        for example if you are doing a sharded evaluation in a distributed environment.

        Args:
            reset_state (bool): If true, clear the data collected so far.

        Returns:
            A dict with keys ``'topk_docs'`` and ``'special_docs'`` which hold the content of attributes of this class with the same name.
        """
        results = dict()
        # deepcopy the results so the changes made by user do not impact the state of this class
        results["topk_docs"] = copy.deepcopy(self.topk_docs)
        results["special_docs"] = copy.deepcopy(self.special_docs)
        if reset_state:
            self.reset_state()
        return results

    def merge_result_dump(self, results: Dict) -> None:
        """Merge the results exported from another instance of this class with their
        :meth:`export_result_dump` method.

        Args:
            results (Dict): The exported results from the other instances of ``ResultHeapq``.
        """
        logger.debug("Merge dump of collected score from other ResultHeapq instances.")

        for qid, scores_ids in results["topk_docs"].items():
            for score, docid in scores_ids:
                self.add_triplet(qid=qid, docid=docid, score=score)

        if "special_docs" in results:
            self.special_docs["qids"].extend(results["special_docs"]["qids"])
            self.special_docs["docids"].extend(results["special_docs"]["docids"])
            self.special_docs["scores"].extend(results["special_docs"]["scores"])

    def get_state_dict(self) -> Dict:
        """Export all the attributes of this class in a dictionary.

        You can use the output of this method to recreate an identical instance of this class later.

        Returns:
            mapping from the name of attributes of this class to their value.
        """
        state_dict = dict()
        state_dict["topk"] = self.topk
        # deepcopy so user changes outside of this class won't corrupt the internal state
        state_dict["special_docids"] = copy.deepcopy(self.special_docids)
        state_dict["special_docs"] = copy.deepcopy(self.special_docs)
        state_dict["topk_docs"] = copy.deepcopy(self.topk_docs)
        return state_dict



import inspect
import json
import os
import pickle as pkl
import time
from pathlib import Path
from typing import Any, Dict, Iterable, List, Optional, Union
from uuid import uuid4

import accelerate
import numpy as np
import torch
import torch.distributed as dist
from tqdm import tqdm

from ..file_utils import JSONLinesWriter
from ..logging_utils import get_logger_with_config
from . import callback_wandb
from .evaluation_args import EvaluationArguments
from .infinite_barrier import InfiniteBarrier

logger, logging_conf = get_logger_with_config("trove")


class RetrievalEvaluatorUtilsMixin:
    def __init__(
        self,
        args: Optional[EvaluationArguments] = None,
        tracker_init_kwargs: Optional[Dict] = None,
        tracker_extra_configs: Optional[Union[List[Dict], Dict]] = None,
        tracker_callbacks: Optional[Union[Any, List[Any]]] = None,
    ) -> None:
        """A simple Mixin for ``RetrievalEvaluator`` that provides convenient utilities.

        Anything that is not part of the main algorithm or is not one of your design decisions should go here.
        For example:

            * dealing with distributed env attributes: is it main process, local process, etc.
            * providing progress bar in distributed environment
            * create a temp file that is shared among processes
            * saving/loading files and checkpoints (what you save and where you save it should be decided in the main class.
              You just define a function here that takes those as arguments and make sure it is written correctly. e.g., only on main process)
            * logging things to wandb, etc.

        But, even small details that impact your design should not be here.
        For example, if there is a temp file that you only delete it under specific circumstances,
        you should check whether to delete the file in the main class.

        Args:
            args (Optional[EvaluationArguments]): Evaluation arguments. Same as what is passed to ``RetrievalEvaluator``.
            tracker_init_kwargs (Optional[Dict]): extra kwargs for initializing experiment trackers.
                Directly passed to tracker init method with dict comprehension.
            tracker_extra_configs (Optional[Union[List[Dict], Dict]]): extra configs to log with experiment trackers.
                These configs are merged with the evaluator and model configs if present.
                You can either pass one config object or a list of config objects.
                Config objects must be instances of python dictionary.
            tracker_callbacks (Optional[Any]): One or multiple custom experiment tracker callbacks.
                Callbacks must have a ``.log()`` method for logging the evaluation metrics.
                and a ``.setup()`` method that is called in the Evaluator constructor to initialize the tracker.
                If in a distributed environment, these methods will only be called from the main global process.
        """
        self.accelerator_state = None
        self._is_distributed = None
        self._rank = None
        self._world_size = None
        self._distributed_info_up_to_date = False
        self.update_distributed_state()

        args._setup_devices

        self.args = args

        # for dist.barrier without timeout
        self._inf_barrier_obj = None
        self.init_infinite_barrier()

        if (
            not self.args.overwrite_output_dir
            and self.args.output_dir is not None
            and Path(self.args.output_dir).exists()
            and list(Path(self.args.output_dir).iterdir())
        ):
            msg = (
                "The given output directory is not empty."
                f" Given output_dir: '{self.args.output_dir}'"
            )
            raise RuntimeError(msg)

        self.tracker_init_kwargs = tracker_init_kwargs
        self.tracker_extra_configs = tracker_extra_configs

        if tracker_callbacks is None:
            self.tracker_callbacks = []
        elif isinstance(tracker_callbacks, list):
            self.tracker_callbacks = [i for i in tracker_callbacks]
        else:
            self.tracker_callbacks = [tracker_callbacks]

        if self.args.report_to is not None and len(self.args.report_to) > 0:
            if len(self.args.report_to) > 1 or self.args.report_to[0] != "wandb":
                other_trackers = [t for t in self.args.report_to if t != "wandb"]
                msg = f"Only 'wandb' experiment tracker is supported. Got: {other_trackers}"
                raise ValueError(msg)
            # If we get to this line, this assert statement is definitely true
            # But, it makes the code easier to read/understand and does not cost anything
            assert len(self.args.report_to) == 1 and self.args.report_to[0] == "wandb"
            self.tracker_callbacks.append(callback_wandb.WandbCallback)

        # Parse arguments to check if we should have a progress bar and echo to stdout in this process
        self.pbar_allowed = self.allowed_on_this_process(self.args.pbar_mode)
        self.print_allowed = self.allowed_on_this_process(self.args.print_mode)

    def update_distributed_state(self) -> None:
        """Set the attributes related to the distributed environment."""
        self.accelerator_state = accelerate.PartialState()
        self._is_distributed = dist.is_initialized()
        if self._is_distributed:
            self._rank = dist.get_rank()
            self._world_size = dist.get_world_size()
        else:
            # If it is not distributed, there is ony one process and that is the main one
            self._rank = 0
            self._world_size = 1

        self._distributed_info_up_to_date = True

    def init_infinite_barrier(self) -> None:
        """Create an instance of ``InfiniteBarrier`` for as a ``dist.barrier()`` alternative
        without timeout."""
        if self.world_size < 2:
            # there is only one process. barrier is not needed
            return
        addr = os.environ["MASTER_ADDR"]
        port = self.get_free_port_on_master()
        self._inf_barrier_obj = InfiniteBarrier(
            master_addr=addr, port=port, world_size=self.world_size, rank=self.rank
        )

    def allowed_on_this_process(self, mode: Optional[str]) -> bool:
        """Parse user provided ``mode`` argument to determine if we are allowed to perform certain
        actions on this processes.

        For example, to check if we should use a progress bar or print print results to stdout.
        If not in a distributed environment, always returns True.
        Valie modes are:

            * ``None`` : all processes are allowed
            * ``one`` : No process is allowed
            * ``main`` : only main process is allowed
            * ``local_main`` : only local main process is allowed
            * ``all`` : all processes are allowed

        Args:
            mode (Optional[str]): mode of operation for a specific action.

        Returns:
            a boolean showing that if we can or cannot perform a specific action from the current process.
        """
        # Check the provided mode is valid
        if mode is not None and mode not in ["none", "main", "local_main", "all"]:
            msg = "Valid values for mode configs (print and pbar) are ['none', 'main', 'local_main', and 'all']"
            raise ValueError(msg)
        # If not distributed, we are the only processes and allowed to do everything
        if not self.is_distributed:
            return True
        if mode is not None and mode == "none":
            return False

        is_allowed = (
            mode is None
            or mode == "all"
            or (mode == "main" and self.is_main_process)
            or (mode == "local_main" and self.is_local_main_process)
        )
        return is_allowed

    def all_gather_object(self, obj: Any) -> List[Any]:
        """Gather a pickleable object from all processes in a distributed environment.

        If not in a distributed environment, it simulates a distributed environment with only one process
        and returns a list with only one item, which is the object from the current (and only) process.

        Args:
            obj (Any): The pickleable object to gather from all processes

        Returns:
            A list of objects. where ``len(objects) == self.world_size``
            and ``objects[rank]`` is the data collected from process with ``rank``
        """
        # We cannot gather None
        assert obj is not None
        if not self.is_distributed:
            # simulate distributed environment
            return [pkl.loads(pkl.dumps(obj))]
        obj_container = [None for _ in range(self.world_size)]
        dist.all_gather_object(obj_container, obj)
        return obj_container

    def gather_object(self, obj: Any, dst: int = 0) -> Optional[List[Any]]:
        """Same as ``all_gather_object()`` but only gathers objects on process with ``rank ==
        dst``.

        Args:
            obj (Any): see :meth:`all_gather_object`
            dst (int): rank of the process that collects the objects from the entire group.

        Returns:
            ``None`` if ``dst != self.rank``. Otherwise, a list of objects. where ``len(objects) == self.world_size``
            and ``objects[rank]`` is the data collected from process with ``rank``
        """
        # We cannot gather None
        assert obj is not None
        if not self.is_distributed:
            # simulate distributed environment
            return [pkl.loads(pkl.dumps(obj))] if self.rank == dst else None
        if self.rank == dst:
            obj_container = [None for _ in range(self.world_size)]
        else:
            obj_container = None
        dist.gather_object(obj=obj, object_gather_list=obj_container, dst=dst)
        return obj_container

    def broadcast_obj(self, obj: Any, src: int = 0) -> Any:
        """Broadcast a pickleable object to all processes.

        Broadcast the obj from process with ``rank == src`` to all processes.
        Returns the broadcasted object.
        I.e., in all processes, returns the object from process with ``rank == src``.

        Args:
            obj (Any): pickleable object to broadcast to all processes.
            src (int): rank of the process that broadcasts the object.

        Returns:
            the broadcasted object from process with ``rank == src``.
        """
        if not self.is_distributed:
            # simulate distributed environment
            return obj
        obj_list = [obj]
        dist.broadcast_object_list(obj_list, src=src)
        return obj_list[0]

    def barrier(self, infinite: bool = False) -> None:
        """Similar to ``dist.barrier()``.

        In future, it should call the appropriate method for different distributed computing
        environments

        Args:
            infinite: If True, use ``InfiniteBarrier`` which can wait forever without a timeout. Otherwise, use regular ``dist.barrier()``.
        """
        if self.world_size < 2:
            return

        if infinite:
            self._inf_barrier_obj.sync()
        else:
            dist.barrier()

    @property
    def device(self) -> torch.device:
        """Current device that model and data should be moved to."""
        return self.args.device

    @property
    def is_distributed(self) -> bool:
        """Returns true if running in a distributed environment."""
        return self._is_distributed

    @property
    def rank(self) -> int:
        """Global rank of current process.

        Returns 0 if not running in distributed environment.
        """
        return self._rank

    @property
    def world_size(self) -> int:
        """The total number of processes.

        Returns 1 if not running in a distributed environment.
        """
        return self._world_size

    @property
    def is_main_process(self) -> bool:
        """Returns true if this is the global main process."""
        return self.args.process_index == 0

    @property
    def is_local_main_process(self) -> bool:
        """Returns true if this is the local main process."""
        return self.args.local_process_index == 0

    def all_devices_are_similar(self) -> bool:
        """Returns true if all devices available in a distributed environment are the same and
        False otherwise."""
        if not self.is_distributed:
            return True
        if self.world_size == 1:
            return True

        # Cuda attributes for current device
        cuda_attrs = torch.cuda.get_device_properties(self.device)
        dev_attrs = dict(
            name=cuda_attrs.name,
            major=cuda_attrs.major,
            minor=cuda_attrs.minor,
            total_memory=cuda_attrs.total_memory,
            multi_processor_count=cuda_attrs.multi_processor_count,
        )
        # Cuda attributes of all devices in the environment
        all_dev_attrs = self.all_gather_object(obj=dev_attrs)

        # Attributes that should be identical to claim similar devices
        for attr in ["name", "major", "minor", "multi_processor_count"]:
            if len(set([d[attr] for d in all_dev_attrs])) != 1:
                return False
        # tolerate up to 1GB of difference in memory. Assume similar memory capacity if difference <=1GB.
        mem_0 = all_dev_attrs[0]["total_memory"]
        for d in all_dev_attrs[1:]:
            if abs(d["total_memory"] - mem_0) >= 2**30:
                return False
        return True

    def get_free_port_on_master(self) -> int:
        """Finds a free tcp port on master host and shares it with all processes."""
        import socket

        def is_port_in_use(port: int) -> bool:
            # returns True if port is in use and False otherwise
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                return s.connect_ex(("localhost", port)) == 0

        port = 0
        if self.is_main_process:
            # we only need to look for the port on the main process
            _sd = (os.getpid() + int(1000 * time.time())) % 2**32
            _rs = np.random.RandomState(_sd)
            while True:
                _port = int(_rs.choice(range(8000, 9999)))
                if not is_port_in_use(_port):
                    port = _port
                    break
        if self.is_distributed:
            # share the port with other processes
            port = self.broadcast_obj(obj=str(port), src=0)
        port = int(port)
        return port

    def get_shared_uuid(self) -> str:
        """Generates a random UUID that is the same across processes in a distributed
        environment."""
        unique_id = str(uuid4().hex)
        if self.is_distributed:
            # If in a distributed environment,
            # use the random name generated by the master processes
            unique_id = self.broadcast_obj(obj=unique_id, src=0)
        return unique_id

    def pbar(self, *args, **kwargs) -> tqdm:
        """Returns a ``tqdm`` instance.

        It is disabled if pbars are not allowed in this process.
        """
        kwargs["disable"] = kwargs.get("disable", False) or not self.pbar_allowed
        kwargs["dynamic_ncols"] = True
        return tqdm(*args, **kwargs)

    def _print(self, *args, **kwargs) -> None:
        """Call the builtin ``print()`` function we are allowed to print to stdout from this
        process."""
        if self.print_allowed:
            print(*args, **kwargs)

    def initialize_trackers(self) -> None:
        """Initialize experiment trackers."""
        if not self.is_main_process:
            return
        if not self.tracker_callbacks:
            return

        _run_name = None
        if (
            self.args.run_name is not None
            and self.args.run_name != self.args.output_dir
        ):
            _run_name = self.args.run_name

        for i in range(len(self.tracker_callbacks)):
            # instantiate trakcker callbacks if they are not already instantiated
            if inspect.isclass(self.tracker_callbacks[i]):
                tracker_cls = self.tracker_callbacks[i]
                self.tracker_callbacks[i] = tracker_cls()
            self.tracker_callbacks[i].setup(
                run_name=_run_name,
                args=self.args,
                model=self.model,
                init_kwargs=self.tracker_init_kwargs,
                extra_configs=self.tracker_extra_configs,
            )

    def log_metrics(self, metrics: Dict) -> None:
        """Log metrics to experiment trackers."""
        if not self.is_main_process:
            return
        if not self.tracker_callbacks:
            return
        for tracker in self.tracker_callbacks:
            tracker.log(metrics)

    def write_json(self, obj: Any, path: os.PathLike, **kwargs) -> None:
        """Write json file only from the main process.

        Args:
            obj: json serializable object to write to json file.
            path: path to destination file.
            **kwargs: keyword arguments passed to ``json.dump()``.
        """
        if not self.is_main_process:
            return
        Path(path).parent.mkdir(exist_ok=True, parents=True)
        with open(path, "w") as f:
            json.dump(obj, f, **kwargs)

    def write_json_lines(
        self, obj_list: Iterable[Any], path: os.PathLike, **kwargs
    ) -> None:
        """Write json lines file only from the main process.

        Args:
            obj_list: list of json serializable objects to write to json file.
            path: path to destination file.
            chunk_size: if provided, force flush the file buffer every ``chunk_size`` records.
            **kwargs: keyword arguments passed to ``json.dumps()``.
        """
        if not self.is_main_process:
            return
        with JSONLinesWriter(path=path, **kwargs) as writer:
            writer.add(obj_list)



import copy
from collections import defaultdict
from typing import Dict, List, Optional, Tuple, Union

import torch

from ..logging_utils import get_logger_with_config

logger, logging_conf = get_logger_with_config("trove")


class FastResultHeapq:
    def __init__(
        self,
        topk: int = 100,
        special_docids: Optional[
            Dict[str, Union[List[str], Dict[str, Union[int, float]]]]
        ] = None,
    ) -> None:
        """keeps track of the topk largest scores for each query.

        This is much faster than the simple python heapq used in :class:`~trove.containers.result_heapq.ResultHeapq`.
        This class requires to iterate over document and query batches in nested for loops,
        and crucially, the **outer loop generates the document batches**. I.e.,

        .. code-block:: python

            for doc_batch in corpus:
                for query_batch in queries:
                    fast_result_heapq(scores, ...)

        It keeps track of the topk most similar documents seen so far. It buffers
        all score batches for each ``doc_batch`` and merges them with results collected
        so far right before moving on to the next ``doc_batch``.

        Unlike ``ResultHeapq``, it uses GPUs for computation if it receives tensors already on GPU.


        This class does not provide all the nice utilities for input/output formatting that ``ResultHeapq`` does.
        To use those utilities, after you are done collecting the similarities, export the data from this class
        and import it into a fresh instance of ``ResultHeapq``.

        The arguments are the same as :class:`~trove.containers.result_heapq.ResultHeapq`. See :class:`~trove.containers.result_heapq.ResultHeapq` docstring for details.
        """
        self.topk = topk

        # Holds topk docs seen so far (except current batch)
        # each element of topk_docs_so_far['indices'] corresponds to one `query_batch`
        # Same for topk_docs_so_far['scores']
        self.topk_docs_so_far = None
        # Same structure as `self.topk_docs_so_far`
        # Collects topk documents from current `doc_batch`
        self.topk_docs_batch = None
        # all docids in the order that they are iterated over
        # It is a flat list of docids.
        self.all_docids = None
        # It is a nested list, where each element is the list of qids for the corresponding `query_batch`
        self.all_qids = None
        # docids for the current 'doc_batch'
        self.current_doc_batch_ids = None
        # mapping from document id to its index in the current 'doc_batch'
        self.current_doc_batch_id2idx = None
        self.special_docs = None
        self.reset_state()

        self.special_docids = defaultdict(set)
        if special_docids is not None:
            for qid in special_docids.keys():
                _docids = special_docids[qid]
                if isinstance(_docids, dict):
                    # special_docids is in nested_qrel format
                    _docids = list(_docids.keys())
                self.special_docids[qid].update(_docids)
        self.special_docids = dict(self.special_docids)

    def export_result_dump(
        self, reset_state: bool = False
    ) -> Dict[str, Dict[str, List[Tuple[str, float]]]]:
        """Export all the results collected so far with the same format as ``ResultHeapq``

        see :class:`~trove.containers.result_heapq.ResultHeapq` for details.
        """
        # Make sure all the collected similarity scores are merged into the final topk documents.
        self.update_topk_records()

        topk_docs = defaultdict(list)
        for q_batch_idx in range(len(self.topk_docs_so_far["indices"])):
            qids = self.all_qids[q_batch_idx]
            topk_scores = self.topk_docs_so_far["scores"][q_batch_idx]
            topk_scores = topk_scores.detach().cpu().tolist()
            topk_indices = self.topk_docs_so_far["indices"][q_batch_idx]
            topk_indices = topk_indices.detach().cpu().tolist()
            topk_docids = [
                [self.all_docids[idx] for idx in row_indices]
                for row_indices in topk_indices
            ]

            for qid, scores, docids in zip(qids, topk_scores, topk_docids):
                topk_docs[qid].extend(zip(scores, docids))
        special_docs = copy.deepcopy(self.special_docs)
        results = {"topk_docs": dict(topk_docs), "special_docs": special_docs}
        if reset_state:
            self.reset_state()
        return results

    def reset_state(self) -> None:
        """Clear the data collected so far."""
        self.topk_docs_so_far = dict(indices=[], scores=[])
        self.topk_docs_batch = dict(indices=[], scores=[])
        self.all_docids = list()
        self.all_qids = list()
        self.special_docs = dict(qids=[], docids=[], scores=[])
        self.current_doc_batch_ids = None
        self.current_doc_batch_id2idx = None

    def update_topk_records(self) -> None:
        """Merges topk records collected so far with topk records from the most recent batch
        selects the new topk records."""
        if not len(self.topk_docs_batch["indices"]):
            # It is the very first query_batch. We have not collected anything yet.
            return

        if not len(self.topk_docs_so_far["indices"]):
            # These are scores for the very doc_batch. There is nothing to merge them with.
            self.topk_docs_so_far["indices"] = self.topk_docs_batch["indices"]
            self.topk_docs_so_far["scores"] = self.topk_docs_batch["scores"]
            self.topk_docs_batch = dict(indices=[], scores=[])
            return

        # Sanity check. They should all have the same number of query batches.
        assert all(
            (
                len(self.topk_docs_so_far["scores"])
                == len(self.topk_docs_so_far["indices"]),
                len(self.topk_docs_batch["indices"])
                == len(self.topk_docs_so_far["indices"]),
                len(self.topk_docs_batch["scores"])
                == len(self.topk_docs_so_far["indices"]),
            )
        )

        for q_batch_idx in range(len(self.topk_docs_so_far["indices"])):
            prev_scores = self.topk_docs_so_far["scores"][q_batch_idx]
            prev_indices = self.topk_docs_so_far["indices"][q_batch_idx]

            batch_scores = self.topk_docs_batch["scores"][q_batch_idx]
            batch_indices = self.topk_docs_batch["indices"][q_batch_idx]

            # update batch index to account for documents seen in previous batches.
            batch_indices = (
                batch_indices + len(self.all_docids) - len(self.current_doc_batch_ids)
            )

            all_scores = torch.concat([prev_scores, batch_scores], dim=1)
            all_indices = torch.concat([prev_indices, batch_indices], dim=1)

            new_topk_scores, tmp_topk_indices = torch.topk(
                all_scores,
                k=min(self.topk, all_scores.shape[1]),
                dim=-1,
                largest=True,
                sorted=True,
            )
            new_topk_indices = all_indices[
                torch.arange(
                    all_indices.shape[0], device=all_indices.device, dtype=torch.long
                )[:, None],
                tmp_topk_indices,
            ]

            self.topk_docs_so_far["scores"][q_batch_idx] = new_topk_scores
            self.topk_docs_so_far["indices"][q_batch_idx] = new_topk_indices

        self.topk_docs_batch = dict(indices=[], scores=[])

    def __call__(
        self,
        scores: torch.Tensor,
        qids: List[str],
        docids: List[str],
        is_first_query_batch: bool,
        **kwargs,
    ) -> None:
        """Adds the calculated scores between a batch of queries and passages.

        Intended to be used as a callable in a pytorch evaluation loop.

        **remember to call :meth:`self.update_topk_records` for the very last manually.**
        It is not done automatically for the last batch.

        Args:
            scores (torch.Tensor): query document similarity scores. Shape: ``[NUM_QUERIES, NUM_DOCS]``
            qids (List[str]): List of query ids in the batch of length ``NUM_QUERIES``
            docids (List[str]): List of document ids in the batch of length ``NUM_DOCS``
            is_first_query_batch (bool): **Must** be set true, if we are processing the very first
                query batch for the current document batch..
            **kwargs: No effect. Just for compatibility with other classes.
        """
        if is_first_query_batch:
            # It is the first query_batch for a new doc_batch.
            # First, merge the results collected for previous 'doc_batch' before processing the new 'doc_batch'
            self.update_topk_records()
            if len(self.special_docids):
                self.current_doc_batch_id2idx = {_id: i for i, _id in enumerate(docids)}
            self.current_doc_batch_ids = docids
            self.all_docids.extend(docids)

        if not len(self.topk_docs_so_far["indices"]):
            # It is the very first 'doc_batch'. qids repeat in the same order for all doc batches.
            # Collect them only once.
            self.all_qids.append(qids)

        topk_scores, topk_indices = torch.topk(
            scores, k=min(self.topk, scores.shape[1]), dim=-1, largest=True, sorted=True
        )
        self.topk_docs_batch["scores"].append(topk_scores)
        self.topk_docs_batch["indices"].append(topk_indices)

        if len(self.special_docids):
            doc_id2idx = self.current_doc_batch_id2idx
            row_indices = list()
            col_indices = list()
            for qidx, qid in enumerate(qids):
                if qid not in self.special_docids:
                    continue
                _cell_docids = list(self.special_docids[qid].intersection(doc_id2idx))
                if not _cell_docids:
                    continue
                _col_indices = [doc_id2idx[_id] for _id in _cell_docids]
                col_indices.extend(_col_indices)
                row_indices.extend([qidx] * len(_cell_docids))
                self.special_docs["qids"].extend([qid] * len(_cell_docids))
                self.special_docs["docids"].extend(_cell_docids)

            if len(row_indices):
                cell_scores = scores[row_indices, col_indices].cpu().tolist()
                self.special_docs["scores"].extend(cell_scores)



import torch
import torch.nn.functional as F
from torch import nn

from ..logging_utils import get_logger_with_config
from .loss_base import RetrievalLoss
from .model_args import ModelArguments

logger, logging_conf = get_logger_with_config("trove")


class InfoNCELoss(RetrievalLoss):
    _alias = "infonce"

    def __init__(self, args: ModelArguments, **kwargs) -> None:
        """Implements InfoNCE loss."""
        super().__init__()

        temp_init_value = float(args.temperature)
        if args.temperature_learnable:
            msg = (
                "In ModelArguments, 'temperature_learnable' is set to True."
                " This feature is incomplete and the learned temperature value is not saved to checkpoints."
                " So you cannot access it later."
            )
            logger.warning(msg)
            _temp_tensor = torch.tensor(temp_init_value)
            if "dtype" in kwargs:
                _temp_tensor = _temp_tensor.to(kwargs["dtype"])
            if "device" in kwargs:
                _temp_tensor = _temp_tensor.to(kwargs["device"])
            self.temperature = nn.Parameter(_temp_tensor)
        else:
            self.temperature = temp_init_value

        reduction = kwargs.get("reduction", "mean")
        self.cross_entropy = nn.CrossEntropyLoss(reduction=reduction)

    def forward(self, logits: torch.Tensor, **kwargs) -> torch.Tensor:
        """Calculates the loss given the similarity scores between query and passages.

        The ``logits`` argument contains the similarity scores between queries and passages for the entire batch.
        For each query in the batch, there are K corresponding passages, where 1 is positive and k-1 are negative passages.
        For each query the positive passages comes first before negative passages, e.g., ``passages_for_q1 = [pos, neg, neg, neg, neg, ..., neg]``

        To use in-batch negatives, the passages for all queries are concatenated in a list.
        I.e., ``all_passages = passages_for_q1 + passages_for_q2 + ... + passages_for_qn``
        As a result, the index of the positive passage for query `i` is `i * K` where `K` is the number of passages per query.

        Args:
            logits (torch.Tensor): The similarity scores between queries and passages. shape: `[NUM_QUERIES, NUM_PASSAGES]`
            **kwargs: Not used. Just to make the signature compatible with other losses.

        Returns:
            the InfoNCE loss value.
        """
        target = torch.arange(logits.size(0), device=logits.device, dtype=torch.long)
        passages_per_query = logits.size(1) // logits.size(0)
        # index of the positive passage in the combined list of documents
        target = target * passages_per_query
        loss = self.cross_entropy(logits / self.temperature, target)
        return loss


class KLDivergenceLoss(RetrievalLoss):
    _alias = "kl"

    def __init__(self, args: ModelArguments, **kwargs) -> None:
        """Implements KL divergence loss."""
        super().__init__()

        temp_init_value = float(args.temperature)
        if args.temperature_learnable:
            msg = (
                "In ModelArguments, 'temperature_learnable' is set to True."
                " This feature is incomplete and the learned temperature value is not saved to checkpoints."
                " So you cannot access it later."
            )
            logger.warning(msg)
            _temp_tensor = torch.tensor(temp_init_value)
            if "dtype" in kwargs:
                _temp_tensor = _temp_tensor.to(kwargs["dtype"])
            if "device" in kwargs:
                _temp_tensor = _temp_tensor.to(kwargs["device"])
            self.temperature = nn.Parameter(_temp_tensor)
        else:
            self.temperature = temp_init_value

    def forward(
        self, logits: torch.Tensor, label: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """Calculates the loss given the similarity scores between query and passages."""
        if label.size(1) != logits.size(1):
            label = torch.block_diag(*torch.chunk(label, label.shape[0]))

        preds = F.log_softmax(logits / self.temperature, dim=1)
        targets = F.log_softmax(label.double(), dim=1)
        loss = F.kl_div(
            input=preds, target=targets, log_target=True, reduction="batchmean"
        )
        return loss



import os
from pathlib import Path
from typing import Callable, List, Optional

import datasets
import numpy as np

from ..logging_utils import get_logger_with_config

logger, logging_conf = get_logger_with_config("trove")

# hold pointer to functions that read the data from file
# keys are the type of data that the function returns
READERS_REGISTRY = {"qrel": [], "qid": [], "record": []}


class FileLoaderNotFoundError(Exception):
    """No loader found for the given file."""


def available_loaders() -> None:
    """Reports what loader functions are available for each output type."""
    print("Available file loaders for each output type.")
    for output_type, loaders in READERS_REGISTRY.items():
        print(" " * 2, f"Output type: <{output_type}>")
        for fn in loaders:
            fn_fullname = fn.__module__ + "." + fn.__name__
            print(" " * 4, fn_fullname)


def register_loader(output: str) -> Callable:
    """A decorator to register functions that read the data from file.

    Later we loop through the registered functions to find one that can load the desired data from a given file.
    All file reader functions should take two keyword arguments: ``'filepath'`` and ``'num_proc'``.

        * ``'filepath'`` is the file that should be loaded.

        * ``'num_proc'`` is the number of processes that the loader can launch in parallel to load and preprocess the data.

    Each reader function should first check if it can load the given file.
    The loader functions should return ``None`` if it cannot read the given file (e.g., if reader can load CSV but the file is in JSON format).

    We go through all the readers until we find one that can load the file.
    So, **make sure the initial check for ability to read the file is fast.**
    In worst case scenario, it is possible that all readers need to do this check before we find the correct loader.

    The expected output for each loader is as following:

        - ``'qrel'`` : instance of huggingface ``datasets.Dataset`` with ``'qid'``, ``'docid'``, and ``'score'`` columns.

            * ``'qid'`` is of type `str` and represents the query id for the record

            * ``'docid'`` is a list of 'str' values (``List[str]``),
              where each item is the id of one related document for this query.

            * ``'score'`` is a list of `int` or `float` (``List[Union[int, float]]``),
              where ``datasets.Dataset[i]['score'][idx]`` is the similarity score betwen
              query ``'qid'`` and document ``datasets.Dataset[i]['docid'][idx]``.

        - ``'qid'`` : a list of query IDs of type `str` (``List[str]``).
          The returned query IDs **MUST** be unique without duplicate query IDs in the list.

        - ``'record'`` : instance of huggingface ``datasets.Dataset``.
          It just needs to load the records in the given file as-is without any further processing.
          It is recommended to keep the records in the same order as they appear in the given file.
          There are no restrictions on what the columns or their data type are.

    Args:
        output (str): The data that the loader returns. Accepted values are ``'qrel'``, ``'qid'``, and ``'record'``.

    Returns:
        A wrapper that registers the given loader function under ``'output'`` key.
    """
    global READERS_REGISTRY
    if output not in READERS_REGISTRY:
        msg = (
            f"'output' of the file reader should one of '{list(READERS_REGISTRY.keys())}'."
            f" Got '{output}'"
        )
        raise ValueError(msg)

    def _registrar(func: Callable) -> Callable:
        # prioritize user-defined functions over the ones defined by the library.
        if "trove" in func.__module__.split("."):
            READERS_REGISTRY[output].append(func)
        else:
            READERS_REGISTRY[output].insert(0, func)
        return func

    return _registrar


def load_qrel(
    filepath: os.PathLike, num_proc: Optional[int] = None
) -> datasets.Dataset:
    """Load qrel triplets from file.

    It loads grouped triplets of (qid, docid, score) from files.
    It loops through all registered qrel readers until it finds one that can load the given file.
    It returns huggingface ``datasets.Dataset`` with ``'qid'``, ``'docid'``, and ``'score'`` columns
    of ``str``, ``List[str]``, and ``List[float]`` dtype, respectively.
    For each record, ``'docid'`` is a list of document ids that are related to ``'qid'``
    and ``'score'`` is the list of similarity score between ``'qid'`` and each document in ``'docid'``.

    Args:
        filepath (os.PathLike): file to read qrels from
        num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.

    Returns:
        huggingface datasets.Dataset with ``'qid'``, ``'docid'``, and ``'score'`` columns of ``str``, ``List[str]``, and ``List[float]`` dtype, respectively.
    """
    global READERS_REGISTRY
    output = None
    for reader_fn in READERS_REGISTRY["qrel"]:
        output = reader_fn(filepath=filepath, num_proc=num_proc)
        if output is not None:
            break

    if output is None:
        msg = f"Could not find a qrel loader function that can load this file:\n{Path(filepath).as_posix()}"
        raise FileLoaderNotFoundError(msg)

    # Since the readers can also be defined by the user, we ensure that the output conforms to
    # the expected structure and data types.

    if not isinstance(output, datasets.Dataset):
        msg = (
            f"The qrel reader function should return an instance of 'datasets.Dataset'."
            f" Got output of type: '{type(output)}'"
        )
        raise TypeError(msg)

    if set(output.column_names) != {"qid", "docid", "score"}:
        msg = (
            f"qrel reader's output should contain exactly three columns: 'qid', 'docid', and 'score'"
            f" Got: {output.column_names}"
        )
        raise ValueError(msg)

    # Check the data type of the columns
    feats = output.features
    if not (
        isinstance(feats["qid"], datasets.Value)
        and feats["qid"].dtype == "string"
        and isinstance(feats["docid"], datasets.Sequence)
        and isinstance(feats["docid"].feature, datasets.Value)
        and feats["docid"].feature.dtype == "string"
        and isinstance(feats["score"], datasets.Sequence)
        and isinstance(feats["score"].feature, datasets.Value)
        and np.issubdtype(feats["score"].feature.pa_type.to_pandas_dtype(), np.number)
    ):
        expected_dtype = {
            "qid": "str",
            "docid": "List[str]",
            "score": "List[Union[int, float]]",
        }
        msg = (
            f"qrel reader's output should have the following schema (or its equivalents in datasets or pyarrow): {expected_dtype}."
            f" Got: {feats}"
        )
        raise TypeError(msg)

    return output


def load_qids(filepath: os.PathLike, num_proc: Optional[int] = None) -> List[str]:
    """Load a list of unique qids from file.

    It loops through all registered qid readers until it finds one that can load the given file.
    If there is no dedicated qid reader that can load this file,
    it assumes it is a file with qrel triplets and returns the list of unique qids in qrel triplets.

    Args:
        filepath (os.PathLike): file to read qids from
        num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.

    Returns:
        A list of unique query IDs (``List[str]``)
    """
    logger.debug("Reading a list of qids")

    global READERS_REGISTRY
    output = None
    for reader_fn in READERS_REGISTRY["qid"]:
        output = reader_fn(filepath=filepath, num_proc=num_proc)
        if output is not None:
            break

    if output is None:
        qrels = load_qrel(filepath=filepath, num_proc=num_proc)
        if qrels is not None:
            if qrels._indices is not None:
                # Dataset is subsampled and the underlying arrow table is not the same as data that dataset uses.
                output = list(set(qrels["qid"]))
            else:
                # Using the underlying pyarrow table is faster/more efficient in some cases
                output = qrels._data.table["qid"].unique().to_pylist()

    if output is None:
        msg = f"Could not find a qid or qrel loader function that can load this file:\n{Path(filepath).as_posix()}"
        raise FileLoaderNotFoundError(msg)

    if len(output) != len(set(output)):
        msg = "The loaded list of qids contains duplicates, which is not allowed."
        raise RuntimeError(msg)

    if not isinstance(output, list):
        msg = f"Expected a output of type 'List'. Got: '{type(output)}'"
        raise TypeError(msg)
    if not isinstance(output[0], str):
        msg = f"qids must be of type 'str'. Got: '{type(output[0])}'"
        raise TypeError(msg)

    return output


def load_records(
    filepath: os.PathLike, num_proc: Optional[int] = None
) -> datasets.Dataset:
    """Load records/rows from file.

    It loops through all registered record readers until it finds one that can load the given file.
    It returns an instance of huggingface ``datasets.Dataset`` that contain the records in the file
    without any further processing or modification.

    Load records/rows from a given file into an instance of ``datasets.Dataset``

    Args:
        filepath (os.PathLike): file to read the records from.
        num_proc (Optional[int]): Max number of processes when reading and pre-processing the data.

    Returns:
        an instance of ``datasets.Dataset`` containing records from the given file.
    """
    global READERS_REGISTRY
    output = None
    for reader_fn in READERS_REGISTRY["record"]:
        output = reader_fn(filepath=filepath, num_proc=num_proc)
        if output is not None:
            break

    if output is None:
        msg = f"Could not find a record loader function that can load this file:\n{Path(filepath).as_posix()}"
        raise FileLoaderNotFoundError(msg)

    if not isinstance(output, datasets.Dataset):
        msg = (
            f"The record reader function should return an instance of 'datasets.Dataset'."
            f" Got output of type: '{type(output)}'"
        )
        raise TypeError(msg)
    return output



import copy
import gc
import itertools
import json
import math
import os
import time
from collections import defaultdict
from contextlib import nullcontext
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import torch
from torch import nn
from torch.utils.data import DataLoader
from transformers import PreTrainedModel, PreTrainedTokenizerBase, default_data_collator

from .. import cache_manager, file_utils
from ..containers import FastResultHeapq, ResultHeapq
from ..data import EncodingDataset, MultiLevelDataset
from ..logging_utils import get_logger_with_config, rpath
from ..modeling import PretrainedRetriever, RetrieverOutput
from . import metric_utils
from .evaluation_args import EvaluationArguments
from .evaluator_mixin_utils import RetrievalEvaluatorUtilsMixin
from .metrics import IRMetrics

logger, logging_conf = get_logger_with_config("trove")


class RetrievalEvaluator(RetrievalEvaluatorUtilsMixin):
    def __init__(
        self,
        args: Optional[EvaluationArguments] = None,
        model: Union[PretrainedRetriever, PreTrainedModel, nn.Module] = None,
        tokenizer: Optional[PreTrainedTokenizerBase] = None,
        data_collator: Optional[Callable] = None,
        eval_dataset: Optional[MultiLevelDataset] = None,
        compute_metrics: Optional[IRMetrics] = None,
        logit_collector: Optional[Union[ResultHeapq, FastResultHeapq]] = None,
        tracker_init_kwargs: Optional[Dict] = None,
        tracker_extra_configs: Optional[Union[List[Dict], Dict]] = None,
        tracker_callbacks: Optional[Union[Any, List[Any]]] = None,
    ) -> None:
        """Simple class for evaluating retrieval performance, mining hard negatives, or just
        computing the embeddings.

        Args:
            args (Optional[EvaluationArguments]): general arguments to control the evaluation/encoding process.
            model (Union[PretrainedRetriever, PreTrainedModel, nn.Module]): retriever model to use
            tokenizer (Optional[PreTrainedTokenizerBase]): (Not used currently)
            data_collator (Optional[Callable]): callable to create a batch from a list of examples.
                It should be able to tokenize the text if embeddings are not precomputed.
            eval_dataset (Optional[MultiLevelDataset]):
                Evaluate the performance on this dataset.
            compute_metrics (Optional[IRMetrics]):
                an instance of IRMetrics class to calculate the IR metrics from predicted scores and ground truth qrels.
            logit_collector (Optional[ResultHeapq]):
                an instance of ResultHeapq that should take the similarity scores for each batch
                and keep the topk most similar documents with their scores for each query.
            tracker_init_kwargs (Optional[Dict]): extra kwargs for initializing experiment trackers. See :class:`~trove.evaluation.evaluator_mixin_utils.RetrievalEvaluatorUtilsMixin` for details.
            tracker_extra_configs (Optional[Union[List[Dict], Dict]]): extra configs to log with experiment trackers. See :class:`~trove.evaluation.evaluator_mixin_utils.RetrievalEvaluatorUtilsMixin` for details.
            tracker_callbacks (Optional[Any]): One or multiple custom experiment tracker callbacks.
                See :class:`~trove.evaluation.evaluator_mixin_utils.RetrievalEvaluatorUtilsMixin` for details.
        """
        super().__init__(
            args=args,
            tracker_init_kwargs=tracker_init_kwargs,
            tracker_extra_configs=tracker_extra_configs,
            tracker_callbacks=tracker_callbacks,
        )

        self.model = model
        self.tokenizer = tokenizer

        if (
            compute_metrics is None
            and self.args.ir_metrics_k_values is not None
            and self.args.ir_metrics_k_values != ""
        ):
            _k_values = [
                int(ks.strip())
                for ks in self.args.ir_metrics_k_values.strip().split(",")
            ]
            _mkw = {}
            if self.args.ir_metrics_relevance_threshold is not None:
                _mkw["relevance_threshold"] = self.args.ir_metrics_relevance_threshold
            compute_metrics = IRMetrics(k_values=_k_values, **_mkw)
            del _mkw
            logger.debug(f"Create IRMetrics with these k_values: '{_k_values}'")
        self.compute_metrics = compute_metrics

        if data_collator is None:
            self.data_collator = default_data_collator
            logger.warning(
                "You are using `default_data_collator`. All your embeddings should be precomputed"
                " It will raise an exception if you try embed text."
            )
        else:
            self.data_collator = data_collator

        if logit_collector is None:
            _topk = None
            if self.args.search_topk is not None:
                _topk = self.args.search_topk
            elif self.compute_metrics is not None:
                _topk = metric_utils.get_measures_max_cutoff(
                    self.compute_metrics.metric_ids
                )
            if _topk is not None:
                logger.debug(f"Create FastResultHeapq with topk: {_topk}")
                logit_collector = FastResultHeapq(topk=_topk)
        self.logit_collector = logit_collector

        self.eval_dataset = eval_dataset

        if self.args.encoding_cache_dir is not None:
            self._shared_cache_uuid = None
            self.cache_pardir = self.args.encoding_cache_dir
        else:
            self._shared_cache_uuid = self.get_shared_uuid()
            self.cache_pardir = cache_manager.get_cache_pardir(
                artifact_content="embedding",
                artifact_type="final",
                cache_uuid=self._shared_cache_uuid,
            )
        logger.debug(
            f"Embeddings are cached in this directory if not specified otherwise: {rpath(self.cache_pardir)}"
        )

        # arguments used to instantiate torch DataLoader for encoding new documents
        self.dataloader_kwargs_encode = dict(
            batch_size=self.args.per_device_eval_batch_size,
            collate_fn=self.data_collator,
            num_workers=self.args.dataloader_num_workers,
            pin_memory=self.args.dataloader_pin_memory,
            persistent_workers=self.args.dataloader_persistent_workers,
            drop_last=False,
            shuffle=False,
            prefetch_factor=self.args.dataloader_prefetch_factor,
        )
        # arguments used to instantiate torch DataLoader reading embeddings from cache and calculating similarities
        self.dataloader_kwargs_matmul = {
            **self.dataloader_kwargs_encode,
            "batch_size": self.args.per_device_matmul_batch_size,
        }

        # In distributed environments, you get the option to shard the dataset based on device throughputs
        # and not evenly for all devices.
        self._shard_weights = None

        self.model = self.model.to(self.device)
        self.model.eval()

        self.initialize_trackers()

    def _forward_pass(self, *args, **kwargs) -> RetrieverOutput:
        """Call model's forward pass with necessary context managers."""
        with (
            torch.cuda.amp.autocast()
            if self.args.fp16 or self.args.bf16
            else nullcontext()
        ):
            with torch.no_grad():
                output = self.model(*args, **kwargs)
        return output

    def get_shard_weights(self, dataset: EncodingDataset) -> Optional[List[float]]:
        """Calculates relative shard sizes based on device performance.

        In each process, it runs the model on a small subset of the given dataset (identical across processes)
        and judges the model throughput based on the time that it takes to process this subset.
        Throughput in each process is used as the shard weight for that processes.

        Args:
            dataset (EncodingDataset): dataset to use for benchmarking. We only use a very small subset of it.

        Returns:
            ``None`` if fair sharding is disabled. A list of shard weights such that ``output[rank]`` is
            the shard weight for the process with ``rank``.
        """
        if not self.args.fair_sharding or not (
            self.is_distributed and self.world_size > 1
        ):
            # fair sharding disabled
            return None
        if self._shard_weights is not None:
            # We only compute them once in each session
            return self._shard_weights

        if self.all_devices_are_similar():
            # devices are the same. Practically, there is no fair sharding here
            # but we set identical shard weights that imply even sharding to avoid calling `self.all_devices_are_similar()` every time
            self._shard_weights = [1.0] * self.world_size
            logger.debug(f"shard weights: {self._shard_weights}")
            return self._shard_weights

        if isinstance(dataset, torch.utils.data.IterableDataset):
            msg = f"Fair sharding is not supported with Iterable datasets. Got: '${type(dataset)}"
            raise ValueError(msg)

        if (
            dataset.shard_idx is not None
            and dataset.shard_idx >= 0
            and dataset.num_shards is not None
            and dataset.num_shards > 0
        ):
            msg = (
                "'dataset' for computing fair shard weights must not be sharded to get"
                " consistent measurements across processes."
                " Got sharded dataset: '{dataset}'."
            )
            raise ValueError(msg)

        logger.info("Run benchmark to calculate shard weights.")

        # the type of data in 'EncodingDataset'
        if dataset.is_query:
            _key = "query"
        else:
            _key = "passage"

        # It is not only the GPU that impacts the performance. So, we choose num_samples for benchmarking
        # such that all dataloader workers have to run at least twice, which reduces the noise.
        num_samples = self.dataloader_kwargs_encode["batch_size"]
        if (
            self.dataloader_kwargs_encode["num_workers"]
            and self.dataloader_kwargs_encode["num_workers"] > 0
        ):
            # We want to have all workers occupied
            num_samples *= self.dataloader_kwargs_encode["num_workers"]
        if (
            self.dataloader_kwargs_encode["prefetch_factor"]
            and self.dataloader_kwargs_encode["prefetch_factor"] > 0
        ):
            # we want to prefetch as many as possible
            num_samples *= self.dataloader_kwargs_encode["prefetch_factor"]
        # double the number of samples to force them fetch new samples in the eval loop so we can measure the time
        num_samples = int(num_samples) * 2
        sample_indices = range(min(num_samples, len(dataset)))
        # repeat the sample indices again. If two processes are running on the same machine
        # It is possible that the first one reads them and they are held in CPU cache and
        # the second process gets a fake high throughput.
        # So we run through the samples twice in each process and just measure the time for the second pass.
        chosen_idx = list(
            itertools.islice(itertools.cycle(sample_indices), num_samples * 2)
        )

        logger.debug(f"Num samples in shard weight benchmark: {len(chosen_idx):,}")

        # we do not want cache
        with dataset.disable_cache():
            data_loader = DataLoader(
                dataset, sampler=chosen_idx, **self.dataloader_kwargs_encode
            )

            # index of the batch where the second pass starts and we start measuring time
            start_batch_idx = math.ceil(len(chosen_idx) / data_loader.batch_size) // 2
            start_time = None

            # It is nice to have them start at the same time. Not sure if it improves the benchmark or not.
            self.barrier()

            for i, batch in enumerate(
                self.pbar(data_loader, desc=f"Fair sharding measurements")
            ):
                if start_time is None and i >= start_batch_idx:
                    # start of the second pass
                    start_time = time.perf_counter()
                self._forward_pass(**{_key: batch[_key].to(self.device)})[_key]
            elapsed = time.perf_counter() - start_time
            # Just multiply it by 1000 so we do not end up with very small number in case the process is very slow
            throughput = 1000.0 / elapsed

        # throughput of all processes
        all_throughputs = self.all_gather_object(obj=throughput)
        self._shard_weights = all_throughputs
        logger.debug(f"shard weights: {self._shard_weights}")
        return self._shard_weights

    def _encode_one(self, encoding_ds: EncodingDataset, display_name: str) -> None:
        """Compute and cache the embeddings for a single ``EncodingDataset``.

        If in distributed environment, it shards the dataset across available processes.

        Args:
            encoding_ds (EncodingDataset): Compute and cache embeddings for this dataset.
            display_name (str): Name to use in console logs and progress bars.
        """
        _name = display_name  # short variable names are cleaner

        logger.info(f"Encoding {_name}")

        encoding_ds.load_cache()
        if encoding_ds.is_cache_available:
            # There is nothing new to do
            logger.debug("Using precomputed embeddings from cache.")
            return

        if not encoding_ds.is_writable:
            msg = (
                "It is not possible to write embeddings to cache,"
                " which means the calculated embeddings are thrown away without any use."
            )
            raise RuntimeError(msg)

        if self.is_distributed and self.world_size > 1:
            # (superficially) shard dataset in a distributed environment
            shard_weights = self.get_shard_weights(dataset=encoding_ds)
            encoding_ds.shard(
                shard_idx=self.rank,
                num_shards=self.world_size,
                shard_weights=shard_weights,
            )

        # the type of data in 'EncodingDataset'
        if encoding_ds.is_query:
            _key = "query"
        else:
            _key = "passage"

        data_loader = DataLoader(encoding_ds, **self.dataloader_kwargs_encode)

        # Make sure all processes know that they are creating a new cache file
        # If one creates a new file before others do the check, they assume corrupted cache and raise exception
        self.barrier()

        with encoding_ds.open_cache_io_streams():
            for batch in self.pbar(data_loader, desc=f"Encode {_name}"):
                emb = self._forward_pass(**{_key: batch[_key].to(self.device)})[_key]
                encoding_ds.cache_records(rec_id=batch["orig_rec_id"], value=emb)

        if self.is_distributed and self.world_size > 1:
            # Make sure all processes are done writing to cache
            self.barrier(infinite=True)  # barrier without timeout
            encoding_ds.unshard()

    def encode(
        self,
        eval_dataset: Optional[
            Union[
                MultiLevelDataset,
                Dict[str, MultiLevelDataset],
                List[MultiLevelDataset],
            ]
        ] = None,
        encoding_dataset: Optional[
            Union[EncodingDataset, Dict[str, EncodingDataset], List[EncodingDataset]]
        ] = None,
        cache_pardir: Optional[os.PathLike] = None,
        display_name: str = "eval",
    ) -> None:
        """Encode texts and cache embeddings.

        * For ``eval_dataset``, the query and corpus files that it uses are encoded.
        * If an ``EncodingDataset`` ends up without any cache filepath, it raises an
          exception (there is probably something wrong if you are just computing
          the embeddings and immediately throwing them away).

        Args:
            eval_dataset (Optional[Union[MultiLevelDataset, Dict, List]]): If given,
                encode the query and corpus files used in this dataset.
            encoding_dataset (Optional[Union[EncodingDataset, Dict, List]]): encode the data generated
                by these ``EncodingDataset`` instances.
            cache_pardir (Optional[os.PathLike]): Save the embedding cache here. The order of priority for
                where cache is saved is as following:

                    * cache file path already attached to EncodingDataset instances
                    * some file in ``cache_pardir`` given to this function
                    * some file in ``EvaluationArguments.encoding_cache_dir`` if provided
                    * no cache is saved (raises an exception)
            display_name (str): Name to use in console logs and progress bars.
                Ideally, it should contain some information about the dataset being encoded.
        """
        _name = display_name  # short variable names are cleaner

        if eval_dataset is not None and encoding_dataset is not None:
            msg = "You can pass either eval_dataset or encoding_dataset but not both."
            raise ValueError(msg)
        if eval_dataset is None and encoding_dataset is None:
            eval_dataset = self.eval_dataset
        if eval_dataset is None and encoding_dataset is None:
            msg = "There is no dataset to encode."
            raise ValueError(msg)

        if cache_pardir is None:
            # These encodings must be persistent,
            # So, use self.args.encoding_cache_pardir instead of self.cache_pardir
            cache_pardir = self.args.encoding_cache_dir

        if eval_dataset is not None:
            if isinstance(eval_dataset, dict):
                for _ds_name, _ds in eval_dataset.items():
                    _kw = dict(display_name=f"{_name}-{_ds_name}")
                    self.encode(eval_dataset=_ds, cache_pardir=cache_pardir, **_kw)
            elif isinstance(eval_dataset, list):
                for _ds_name, _ds in enumerate(eval_dataset):
                    _kw = dict(display_name=f"{_name}-{_ds_name}")
                    self.encode(eval_dataset=_ds, cache_pardir=cache_pardir, **_kw)
            else:
                query_ds, corpus_ds = eval_dataset.get_encoding_datasets(
                    encoding_cache_pardir=cache_pardir, load_cache_on_init=False
                )
                for _ds_idx, eds in enumerate(corpus_ds):
                    _kw = dict(display_name=f"{_name}-C{_ds_idx+1}-of-{len(corpus_ds)}")
                    self.encode(encoding_dataset=eds, cache_pardir=cache_pardir, **_kw)
                for _ds_idx, eds in enumerate(query_ds):
                    _kw = dict(display_name=f"{_name}-Q{_ds_idx+1}-of-{len(query_ds)}")
                    self.encode(encoding_dataset=eds, cache_pardir=cache_pardir, **_kw)
        else:
            if isinstance(encoding_dataset, dict):
                for _ds_name, _ds in encoding_dataset.items():
                    _kw = dict(display_name=f"{_name}-{_ds_name}")
                    self.encode(encoding_dataset=_ds, cache_pardir=cache_pardir, **_kw)
            elif isinstance(encoding_dataset, list):
                for _ds_name, _ds in enumerate(encoding_dataset):
                    _kw = dict(display_name=f"{_name}-{_ds_name}")
                    self.encode(encoding_dataset=_ds, cache_pardir=cache_pardir, **_kw)
            else:
                if encoding_dataset.cache_file_name is None:
                    # Create a unique cache file based on the data filename
                    assert cache_pardir is not None
                    _newfile = cache_manager.get_cache_dir(
                        input_data=encoding_dataset.filepath,
                        cache_pardir=cache_pardir,
                        cache_uuid=self._shared_cache_uuid,
                        write_metadata=self.rank == 0,
                    )
                    _newfile = _newfile.joinpath("embedding_cache.arrow").as_posix()
                    encoding_dataset.update_cache_file_name(_newfile, load=False)
                    logger.debug(f"New embedding cache file to set: {rpath(_newfile)}")
                self._encode_one(encoding_dataset, display_name=_name)
                # Free memory. We do not need the cache.
                encoding_dataset.unload_cache()

    def nearest_neighbor_search(
        self,
        query_dataset: Union[EncodingDataset, List[EncodingDataset]],
        corpus_dataset: Union[EncodingDataset, List[EncodingDataset]],
        logit_collector: Union[ResultHeapq, FastResultHeapq],
        cache_pardir: os.PathLike,
        display_name: str,
    ) -> Optional[Dict[str, Dict[str, float]]]:
        """Find the nearest neighbors for each query.

        .. note::

            To save memory in distributed environments, this method only returns the output in process with ``rank == 0`` and returns ``None`` in other processes.

        Args:
            query_dataset (Union[EncodingDataset, List[EncodingDataset]]): One or multiple datasets holding
                the search queries.
            corpus_dataset (Union[EncodingDataset, List[EncodingDataset]]): One or multiple datasets holding
                the documents to search.
            logit_collector (Union[ResultHeapq, FastResultHeapq]): Instance of ``ResultHeapq`` or ``FastResultHeapq`` that takes the scores
                for each batch and keeps the topk most similar documents and their scores for each query.
            cache_pardir (os.PathLike): Write the embedding cache files to this directory.
            display_name (str): Name to use in console logs and progress bars.

        Returns:
            a mapping containing the collected topk similarities, which is the output of the ``logit_collector``.
        """
        _name = display_name  # short variable names are cleaner
        logger.info(f"Nearest neighbor search in: {_name}")

        assert logit_collector is not None
        # Make sure to start from a clean state and remove any scores from previous runs
        logit_collector.reset_state()

        query_datasets = (
            query_dataset if isinstance(query_dataset, List) else [query_dataset]
        )
        corpus_datasets = (
            corpus_dataset if isinstance(corpus_dataset, List) else [corpus_dataset]
        )

        for eds in query_datasets + corpus_datasets:
            if eds.cache_file_name is None:
                _newfile = cache_manager.get_cache_dir(
                    input_data=eds.filepath,
                    cache_pardir=cache_pardir,
                    cache_uuid=self._shared_cache_uuid,
                    write_metadata=self.rank == 0,
                )
                _newfile = _newfile.joinpath("embedding_cache.arrow").as_posix()
                eds.update_cache_file_name(file_name=_newfile, load=False)
                logger.debug(f"New embedding cache file to set: {rpath(_newfile)}")

        logger.info("Precompute query embeddings before calculating scores.")
        for i, eds in enumerate(query_datasets):
            self._encode_one(
                eds, display_name=f"{_name}-Q{i+1}-of-{len(query_datasets)}"
            )

        if self.args.precompute_corpus_embs:
            logger.info("Precompute corpus embeddings before calculating scores.")
            for i, eds in enumerate(corpus_datasets):
                self._encode_one(
                    eds, display_name=f"{_name}-C{i+1}-of-{len(corpus_datasets)}"
                )

        dataloader_kwargs_matmul = copy.deepcopy(self.dataloader_kwargs_matmul)
        dataloader_kwargs_encode = copy.deepcopy(self.dataloader_kwargs_encode)
        if self.args.precompute_corpus_embs:
            # If embeddings are precomputed, swap batch sizes of inner and outer loop
            # to speed up topk score calculations
            dataloader_kwargs_encode["batch_size"] = (
                self.args.per_device_matmul_batch_size
            )
            dataloader_kwargs_matmul["batch_size"] = (
                self.args.per_device_eval_batch_size
            )

        query_dataloaders = list()
        for qds in query_datasets:
            # Make sure cache is loaded. So we do not compute them twice or get errors
            qds.load_cache()
            # pre-fetch the query embeddings. It makes it much faster.
            query_dataloaders.append(
                [b for b in DataLoader(qds, **dataloader_kwargs_matmul)]
            )
            qds.unload_cache()  # free some memory
        gc.collect()  # force release memory

        # In the following we will loop over all query document pairs.
        # It seems complicated but it is basically doing this
        # for corpus_file in all_corpuses:
        #   for doc_batch in corpus_file:
        #       for query_file in all_query_files:
        #           for query_batch in query_file:
        #               calc sim(query_batch, doc_batch)

        for corpus_ds in self.pbar(
            corpus_datasets,
            desc=f"{_name} corpus files",
            disable=len(corpus_datasets) < 2,
        ):
            # Make sure cache is loaded. So we do not compute them twice or get errors
            corpus_ds.load_cache()
            if self.is_distributed and self.world_size > 1:
                # (superficially) shard dataset in a distributed environment
                shard_weights = self.get_shard_weights(dataset=corpus_ds)
                corpus_ds.shard(
                    shard_idx=self.rank,
                    num_shards=self.world_size,
                    shard_weights=shard_weights,
                )
            data_loader = DataLoader(corpus_ds, **dataloader_kwargs_encode)

            # Make sure all processes know that they are creating a new cache file
            # If one creates a new file before others do the check, they assume corrupted cache and raise exception
            self.barrier()
            with corpus_ds.open_cache_io_streams():
                for doc_batch in self.pbar(
                    data_loader, desc=f"{_name} docs", leave=len(corpus_datasets) < 2
                ):
                    if "embedding" in doc_batch:
                        doc_emb = doc_batch["embedding"].to(self.device)
                    else:
                        doc_emb = self._forward_pass(
                            **{"passage": doc_batch["passage"].to(self.device)}
                        )["passage"]
                        corpus_ds.cache_records(
                            rec_id=doc_batch["orig_rec_id"], value=doc_emb
                        )
                    # Marks the very first query batch that we process for each doc_batch
                    _is_first_query_batch = True
                    for q_loader in self.pbar(
                        query_dataloaders,
                        desc=f"{_name} query files",
                        disable=len(query_dataloaders) < 2,
                        leave=False,
                    ):
                        for q_batch in self.pbar(
                            q_loader,
                            desc=f"{_name} queries",
                            disable=len(q_loader) < 2,
                            leave=False,
                        ):
                            if "embedding" in q_batch:
                                q_emb = q_batch["embedding"].to(self.device)
                            else:
                                q_emb = self._forward_pass(
                                    **{"query": q_batch["query"].to(self.device)}
                                )["query"]

                            scores = self.model.similarity_fn(
                                query=q_emb, passage=doc_emb
                            )
                            logit_collector(
                                scores=scores,
                                qids=q_batch["rec_id"],
                                docids=doc_batch["rec_id"],
                                is_first_query_batch=_is_first_query_batch,
                            )
                            _is_first_query_batch = False

            if self.is_distributed and self.world_size > 1:
                # Make sure all processes are done writing to cache
                self.barrier()
                corpus_ds.unshard()
            # free up some memory
            corpus_ds.unload_cache()

        # free up some memory
        del doc_batch
        del doc_emb
        del q_loader
        del q_batch
        del q_emb
        del query_dataloaders
        gc.collect()

        if isinstance(logit_collector, FastResultHeapq):
            # Make sure to process the very last batch too.
            logit_collector.update_topk_records()
        logger.debug("Finished calculating scores.")

        if (self.is_distributed and self.world_size > 1) or not isinstance(
            logit_collector, ResultHeapq
        ):
            # Gather topk scores from all processes
            # Also make sure we are working with an instance of ResultHeapq since we later rely on import/export utilities that it provides
            logger.debug(
                "Export logit_collector dump to merge results from all processes and convert from 'FastResultHeapq' to 'ResultHeapq'"
            )
            collected_res = logit_collector.export_result_dump(reset_state=True)
            logger.info("Gathering scores from all processes.")
            print("Gathering scores from all processes.")
            all_collected_res = self.gather_object(obj=collected_res, dst=0)
            del collected_res  # save some memory
            gc.collect()

            if self.is_main_process:
                _logit_collector = ResultHeapq(
                    topk=logit_collector.topk,
                    special_docids=copy.deepcopy(logit_collector.special_docids),
                )
                del logit_collector  # save some memory

                while all_collected_res:
                    print("Merge scores from other devices.")
                    _logit_collector.merge_result_dump(all_collected_res.pop())

                collected_res = _logit_collector.as_qrel_nested_dict(collection="all")
                # save some memory
                _logit_collector.reset_state()
                del _logit_collector
            else:
                # save some memory
                logit_collector.reset_state()
                del logit_collector
                collected_res = None
                gc.collect()
        else:
            # There is only one process and we are working with an instance of ResultHeapq. so, no extra action needed
            collected_res = logit_collector.as_qrel_nested_dict(collection="all")
            logit_collector.reset_state()

        self.barrier(infinite=True)
        return collected_res

    def _evaluate_one(
        self,
        eval_dataset: MultiLevelDataset,
        logit_collector: Union[ResultHeapq, FastResultHeapq],
        cache_pardir: os.PathLike,
        display_name: str,
    ) -> Optional[Dict[str, Union[Any, Dict[str, float]]]]:
        """Run evaluation for a single dataset and return metrics and collected logits (i.e.,
        scores).

        .. note::

            To save memory in distributed environments, this method only returns the output in process with ``rank == 0`` and returns ``None`` in other processes.

        The intermediate embeddings are written to a temporary cache. If the user does not explicitly
        ask to cache the embeddings, the temporary cache is deleted before returning from this function.

        Args:
            eval_dataset (MultiLevelDataset): dataset to evaluate.
            logit_collector (Union[ResultHeapq, FastResultHeapq]): Instance of ``ResultHeapq`` or ``FastResultHeapq`` that takes the scores
                for each batch and keeps the topk most similar documents and their scores for each query.
            cache_pardir (os.PathLike): Write the embedding cache files to this directory.
            display_name (str): Name to use in console logs and progress bars.

        Returns:
            A mapping with two keys. ``metrics`` is the computed metrics; and
            ``logits`` is the subset of similarity scores collected by logit_collector.
        """
        logger.info(f"Evaluating: {display_name}")

        _metrics_max_cutoff = metric_utils.get_measures_max_cutoff(
            self.compute_metrics.metric_ids
        )
        if _metrics_max_cutoff > logit_collector.topk:
            msg = (
                "Number of retrieved document (i.e., 'logit_collector.topk') must be >= max cutoff value of metrics."
                f" Got number of retrieved documents: {logit_collector.topk} and metrics max cutoff value: {_metrics_max_cutoff}"
            )
            raise RuntimeError(msg)

        query_datasets, corpus_datasets = eval_dataset.get_encoding_datasets(
            encoding_cache_pardir=cache_pardir, load_cache_on_init=False
        )
        search_results = self.nearest_neighbor_search(
            query_dataset=query_datasets,
            corpus_dataset=corpus_datasets,
            logit_collector=logit_collector,
            cache_pardir=cache_pardir,
            display_name=display_name,
        )
        # free up some memory
        del query_datasets
        del corpus_datasets
        gc.collect()

        if not self.is_main_process:
            # to save memory, we only calculate the output and return it in the main process.
            # And avoid duplicate computation in other processes
            return None

        logger.debug("Prep Qrel and compute Metrics.")
        # Ground truth qrel
        gt_qrel = eval_dataset.get_qrel_nested_dict(return_global_ids=True)
        metrics = self.compute_metrics.compute(
            scores=search_results["topk_docs"], qrels=gt_qrel
        )

        output = {"metrics": metrics, "logits": search_results}
        return output

    @file_utils.call_cleanup("remove_temp_cache_pardir")
    def evaluate(
        self,
        eval_dataset: Optional[
            Union[MultiLevelDataset, Dict[str, MultiLevelDataset]]
        ] = None,
        logit_collector: Optional[
            Union[
                Union[ResultHeapq, FastResultHeapq],
                Dict[str, Union[ResultHeapq, FastResultHeapq]],
            ]
        ] = None,
        cache_pardir: Optional[os.PathLike] = None,
        display_name: str = "eval",
        broadcast_output: Optional[bool] = None,
    ) -> Optional[
        Union[
            Dict[str, Union[Any, Dict[str, float]]],
            Dict[str, Dict[str, Union[Any, Dict[str, float]]]],
        ]
    ]:
        """Run evaluation and return metrics and collected logits (i.e., scores).

        The intermediate embeddings are written to a temporary cache. If the user does not explicitly
        ask to cache the embeddings, the temporary cache is deleted before returning from this function.

        Args:
            eval_dataset (Optional[): dataset to evaluate (if not provided, use ``RetrievalEvaluator.eval_dataset``)
            logit_collector (Optional[Union[Union[ResultHeapq, FastResultHeapq], Dict[str, Union[ResultHeapq, FastResultHeapq]]]]): One or multiple instances
                of ``ResultHeapq`` or ``FastResultHeapq`` that take the scores for each batch
                and keep track of the topk most similar documents for each query.
            cache_pardir (Optional[os.PathLike]): Write the embedding cache files to this
                directory (if not provided, use ``RetrievalEvaluator.cache_pardir``)
            display_name (str): Name to use in console logs and progress bars.
            broadcast_output (Optional[bool]): (only for distributed environments) If true, the output is duplicated
                across all processes (i.e., this method returns identical output in all processes).
                If False, only the main process returns the output and other processes return None.
                Set it to False to save memory on machines with multiple GPUs.

        Returns:
            a mapping with two keys, ``metrics`` and ``logits``. ``metrics`` is a mapping from metric name to metric value.
            ``logits`` is the subset of scores collected by ``logit_collector`` and
            is obtained by calling ``logit_collector.as_qrel_nested_dict``.
            If ``eval_dataset`` is a dict, we return a mapping from keys in ``eval_dataset``
            to the results for the corresponding dataset.
        """
        _name = display_name  # short variable names are cleaner

        if broadcast_output is None:
            broadcast_output = self.args.broadcast_output

        if eval_dataset is None:
            eval_dataset = self.eval_dataset
        if eval_dataset is None:
            msg = "There is no dataset to evaluate."
            raise ValueError(msg)
        if logit_collector is None:
            logit_collector = self.logit_collector
        if cache_pardir is None:
            cache_pardir = self.cache_pardir

        if isinstance(eval_dataset, dict):
            results = dict()
            for ds_name, ds in eval_dataset.items():
                if isinstance(logit_collector, dict):
                    lc = logit_collector[ds_name]
                else:
                    lc = logit_collector
                _out = self._evaluate_one(
                    eval_dataset=ds,
                    logit_collector=lc,
                    cache_pardir=cache_pardir,
                    display_name=f"{_name}-{ds_name}",
                )
                results[ds_name] = _out

        else:
            _results = self._evaluate_one(
                eval_dataset=eval_dataset,
                logit_collector=logit_collector,
                cache_pardir=cache_pardir,
                display_name=_name,
            )
            results = dict(eval=_results)

        if self.is_main_process:
            # reformat the metrics to metrics[ds_name/metric_name] = metric_value
            # This creates a new section for each dataset in wandb dashboard
            metrics_to_log = dict()
            metrics_to_print = dict()
            for ds_name, ds_output in results.items():
                metrics_to_print[ds_name] = ds_output["metrics"]
                for mname, mval in ds_output["metrics"].items():
                    metrics_to_log[f"{ds_name}/{mname}"] = mval

            self._print(json.dumps(metrics_to_print, indent=2))
            self.log_metrics(metrics_to_log)

            if self.args.output_dir is not None:
                logger.info("Writing evaluation results to disk.")
                for ds_name, ds_results in results.items():
                    dst_dir = Path(self.args.output_dir, "evaluation_results")
                    if isinstance(eval_dataset, dict):
                        # only create a subdir for dataset if eval_dataset is a dict and has a corresponding key for each dataset
                        dst_dir = dst_dir.joinpath(ds_name)
                    self.write_json(
                        obj=ds_results["metrics"],
                        path=dst_dir.joinpath("metrics.json"),
                        indent=2,
                    )
                    if self.args.save_eval_topk_logits:
                        for res_k in ds_results["logits"].keys():
                            res_v = ds_results["logits"][res_k]

                            res_local_id = defaultdict(dict)
                            for g_qid, _topdocs in res_v.items():
                                _qid, _ = g_qid.rsplit("_", maxsplit=1)
                                for g_docid, _s in _topdocs.items():
                                    _docid, _ = g_docid.rsplit("_", maxsplit=1)
                                    if _docid in res_local_id[_qid]:
                                        raise RuntimeError(
                                            "Found two records with the same ID. Try setting `save_eval_topk_logits` to `False`"
                                        )
                                    res_local_id[_qid][_docid] = _s
                            res_v = dict(res_local_id)
                            del res_local_id
                            if self.args.output_qrel_format == "tsv":
                                res_path = dst_dir.joinpath(f"logits_{res_k}.tsv")
                            elif self.args.output_qrel_format == "grouped":
                                res_path = dst_dir.joinpath(f"logits_{res_k}.jsonl")
                            else:
                                raise ValueError

                            with file_utils.atomic_write(
                                file=res_path, root="parent"
                            ) as tfile:
                                file_utils.write_qrels(
                                    qrels=res_v,
                                    filepath=tfile,
                                    format=self.args.output_qrel_format,
                                )

            if not isinstance(eval_dataset, dict):
                # If original dataset was not a dict, you should not return the results as dict
                results = list(results.values())[0]
        else:
            results = None

        self.barrier()
        if broadcast_output:
            # copy the output from process 0 to all other processes
            logger.info("Broadcast output to all processes.")
            results = self.broadcast_obj(obj=results, src=0)
        return results

    def _mine_hard_negatives_one(
        self,
        eval_dataset: MultiLevelDataset,
        logit_collector: Union[ResultHeapq, FastResultHeapq],
        cache_pardir: os.PathLike,
        display_name: str,
        num_negs: Optional[int] = None,
    ) -> Optional[List[Dict[str, Union[os.PathLike, Dict[str, Dict[str, float]]]]]]:
        """Mine hard negatives for one eval dataset.

        .. note::

            To save memory in distributed environments, this method only returns the output in process with ``rank == 0`` and returns ``None`` in other processes.

        See docstring for :meth:`mine_hard_negatives` for details.
        """
        logger.info(f"Mining hard negatives for: {display_name}")

        if self.args.no_annot_in_mined_hn:
            gt_qrel = eval_dataset.get_qrel_nested_dict(return_global_ids=True)
        else:
            # pretending to not have any annotations.
            gt_qrel = dict()

        max_gt_docs_per_query = None
        if len(gt_qrel) != 0:
            # maximum number of annotated docs per query
            max_gt_docs_per_query = max([len(item) for item in gt_qrel.values()])

        # free up memory for NN search. We recreate it again later
        del gt_qrel

        if num_negs is None:
            num_negs = logit_collector.topk

        # We want to choose num_negs unannotated docs for each query
        # Instead of excluding annotated docs during NN search (which is slow),
        # We choose num_negs + max_gt_docs_per_query documents for each query to make sure
        # at least num_negs of chosen documents are unannotated
        orig_topk = logit_collector.topk
        if max_gt_docs_per_query is None:
            logit_collector.topk = num_negs
        else:
            logit_collector.topk = num_negs + max_gt_docs_per_query
        # after changing topk, clear the internal state of logit_collector to make sure
        # it holds valid data at all times
        logit_collector.reset_state()

        query_datasets, corpus_datasets = eval_dataset.get_encoding_datasets(
            encoding_cache_pardir=cache_pardir, load_cache_on_init=False
        )
        # unique global file ids that extend query and corpus IDs. See EncodingDataset.__getitem__ for details
        global_id_to_filepath = dict()
        for eds in query_datasets + corpus_datasets:
            global_id_to_filepath[eds.global_id_suffix[1:]] = eds.filepath

        # currently, we can only convert global IDs to local IDs if the ID suffix does not
        # contains '_' character
        if not all(["_" not in _id for _id in global_id_to_filepath.keys()]):
            msg = (
                "At least one of the global file IDs contain an '_' character, "
                "which is not supported at this point."
            )
            raise RuntimeError(msg)

        search_results = self.nearest_neighbor_search(
            query_dataset=query_datasets,
            corpus_dataset=corpus_datasets,
            logit_collector=logit_collector,
            cache_pardir=cache_pardir,
            display_name=display_name,
        )
        # free up some memory
        del query_datasets
        del corpus_datasets
        del eds
        gc.collect()

        if not self.is_main_process:
            # to save memory, we only calculate the output and return it in the main process.
            # And avoid duplicate computation in other processes
            return None
        search_results = search_results["topk_docs"]
        # reverse the topk property of the logit_collector and reset its state to make sure
        # this object remains valid for others to use
        logit_collector.topk = orig_topk
        logit_collector.reset_state()
        gc.collect()

        if self.args.no_annot_in_mined_hn:
            gt_qrel = eval_dataset.get_qrel_nested_dict(return_global_ids=True)
        else:
            # pretending to not have any annotations.
            gt_qrel = dict()

        # mined_qrels[query_file_id][corpus_file_id] contains a subset of mined qrel triplets
        # that their query and documents come from files with 'query_file_id' and 'corpus_file_id' global IDs
        # mined_qrels[query_file_id][corpus_file_id][qid][docid] = similarity_score(qid, docid)
        mined_qrels = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))

        for g_qid, docdata in self.pbar(
            search_results.items(),
            desc="Filter annotated docs from NN search",
            disable=not logging_conf.is_debug(),
        ):
            if len(gt_qrel) == 0:
                # There is no information about annotated docs for any of the queries
                # So all documents are used during HN mining
                _docdata = docdata.items()
            elif g_qid not in gt_qrel:
                # We have annotation information in general but we don't have any
                # information about annotated docs for this specific querys. So we skip it
                continue
            else:
                # exclude annotated docs
                _docdata = [i for i in docdata.items() if i[0] not in gt_qrel[g_qid]]
            _docdata = sorted(_docdata, key=lambda x: (x[1], x[0]), reverse=True)
            _docdata = _docdata[:num_negs]
            # convert global IDs to local IDs
            qid, query_f_id = g_qid.rsplit("_", maxsplit=1)
            for g_docid, docscore in _docdata:
                # convert global IDs to local IDs
                docid, corpus_f_id = g_docid.rsplit("_", maxsplit=1)
                mined_qrels[query_f_id][corpus_f_id][qid][docid] = docscore

        del gt_qrel  # free memory

        logger.debug(
            "Changing mined qrel format to a list of nested qrel dicts with one entry per query-corpus file pair."
        )
        mined_qrels_list = list()
        for query_f_id, query_f_data in mined_qrels.items():
            for corpus_f_id, query_corpus_f_qrels in query_f_data.items():
                mined_qrels_list.append(
                    {
                        "query_file": global_id_to_filepath[query_f_id],
                        "corpus_file": global_id_to_filepath[corpus_f_id],
                        "qrel": dict(query_corpus_f_qrels),
                    }
                )
        if self.args.merge_mined_qrels and len(mined_qrels_list) > 1:
            merged_qrels = defaultdict(dict)
            for pair_data in mined_qrels_list:
                for qid, topdocs in pair_data["qrel"].items():
                    for docid, score in topdocs.items():
                        if docid in merged_qrels[qid]:
                            raise RuntimeError(
                                "Found two records with the same ID. Try setting `merge_mined_qrels` to `False`"
                            )
                        merged_qrels[qid][docid] = score
            mined_qrels_list = [
                {"query_file": "all", "corpus_file": "all", "qrel": dict(merged_qrels)}
            ]

        return mined_qrels_list

    @file_utils.call_cleanup("remove_temp_cache_pardir")
    def mine_hard_negatives(
        self,
        eval_dataset: Optional[
            Union[MultiLevelDataset, Dict[str, MultiLevelDataset]]
        ] = None,
        query_filepath: Optional[Union[os.PathLike, List[os.PathLike]]] = None,
        corpus_filepath: Optional[Union[os.PathLike, List[os.PathLike]]] = None,
        logit_collector: Optional[
            Union[
                Union[ResultHeapq, FastResultHeapq],
                Dict[str, Union[ResultHeapq, FastResultHeapq]],
            ]
        ] = None,
        cache_pardir: Optional[os.PathLike] = None,
        display_name: str = "eval",
        num_negs: Optional[int] = None,
        broadcast_output: Optional[bool] = None,
    ) -> Optional[
        Union[
            List[Dict[str, Union[os.PathLike, Dict[str, Dict[str, float]]]]],
            Dict[str, List[Dict[str, Union[os.PathLike, Dict[str, Dict[str, float]]]]]],
        ]
    ]:
        """Mine the most similar documents for each query as hard negatives.

        Retrieves the topk most similar documents for each query as hard negatives.

        In the case that the resulting ``eval_dataset`` object (whether given or created
        from ``query_filepath``, and ``corpus_filepath``) contains a valid set of qrel triplets,
        queries that have no corresponding qrel triplet are ignored. I.e., no documents are mined for them
        and they are not included in the returned results.

        Mined hard negatives are returned in nested qrel format. But, instead of
        returning one qrel object, it creates one qrel object for each pair of query
        and corpus files (this allows us to read queries and documents from multiple
        files with potentially none-unique IDs across files). It returns a list of dicts, each corresponding to a
        pair of query and corpus files::

            [
                {
                    'query_file': 'path to file that contains the corresponding queries.',
                    'corpus_file': 'path to file that contains the corresponding documents.',
                    'qrel': '''a subset of mined hard negatives that contain only queries and documents
                        from 'query_file' and 'corpus_file'. it is in nested qrel format (i.e., qrel[qid][docid]=similarity(qid, docid)).'''
                }

                ...
            ]

        If ``args.output_dir`` is provided, mined hard negatives are also written to disk in grouped qrel format
        in a json lines file.

        Args:
            eval_dataset: dataset to mine hard negatives for.
            query_filepath: file to read queries from.
            corpus_filepath: file to read documents from.
            logit_collector: Instance of ``ResultHeapq`` or ``FastResultHeapq`` that takes the scores
                for each batch and keeps the topk most similar documents and their scores for each query.
            cache_pardir: Write the embedding cache files to this directory.
            display_name: Name to use in console logs and progress bars.
            num_negs: number of hard negatives to mine per query. If not provided, use the value of ``args.search_topk``.
            broadcast_output (Optional[bool]): (only for distributed environments) If true, the output is duplicated
                across all processes (i.e., this method returns identical output in all processes).
                If false, only the main process returns the output and other processes return ``None``.
                Set it to false to save memory on machines with multiple GPUs.

        Returns:
            A list of mined hard negatives with one entry for each contributing pair of query
            and corpus files. See extended method docstring for details. If ``eval_dataset`` is a dict,
            we return a mapping from keys in ``eval_dataset`` to the described results for
            the corresponding dataset.
        """
        _name = display_name  # short variable names are cleaner

        if (query_filepath is None) != (corpus_filepath is None):
            msg = (
                "You should provide either both or none of 'query_filepath' and 'corpus_filepath'."
                f" Got: query_filepath={query_filepath}\n"
                f"corpus_filepath={corpus_filepath}"
            )
            raise ValueError(msg)

        if eval_dataset is not None and query_filepath is not None:
            msg = "You should provide either eval_dataset or query/corpus file paths. But, not both."
            raise ValueError(msg)

        if query_filepath is not None or corpus_filepath is not None:
            msg = (
                "Mining hard negatives from corpus and query files is not yet supported."
                " Pass an instance of 'MultiLevelDataset' to `eval_dataset` instead."
            )
            raise NotImplementedError(msg)

        if broadcast_output is None:
            broadcast_output = self.args.broadcast_output

        if eval_dataset is None:
            eval_dataset = self.eval_dataset
        if eval_dataset is None:
            msg = "There is no dataset for hard negative mining."
            raise ValueError(msg)
        if logit_collector is None:
            logit_collector = self.logit_collector
        if cache_pardir is None:
            cache_pardir = self.cache_pardir

        if isinstance(eval_dataset, dict):
            results = dict()
            for ds_name, ds in eval_dataset.items():
                if isinstance(logit_collector, dict):
                    lc = logit_collector[ds_name]
                else:
                    lc = logit_collector
                _out = self._mine_hard_negatives_one(
                    eval_dataset=ds,
                    logit_collector=lc,
                    cache_pardir=cache_pardir,
                    display_name=f"{_name}-{ds_name}",
                    num_negs=num_negs,
                )
                results[ds_name] = _out
        else:
            _results = self._mine_hard_negatives_one(
                eval_dataset=eval_dataset,
                logit_collector=logit_collector,
                cache_pardir=cache_pardir,
                display_name=_name,
                num_negs=num_negs,
            )

            results = dict(eval=_results)

        if self.is_main_process:
            if self.args.output_dir is not None:
                logger.info("Writing mined qrel triplets to disk.")
                for ds_name, ds_results in results.items():
                    for pair_idx, pair_data in enumerate(ds_results):
                        dst_dir = Path(self.args.output_dir, "mined_hard_negatives")
                        if isinstance(eval_dataset, dict):
                            # only add ds_name if eval_dataset is a dict and has a corresponding key and NOT if we set ds_name to eval ourselves above
                            dst_dir = dst_dir / ds_name
                        if not self.args.merge_mined_qrels:
                            dst_dir = dst_dir / f"query_corpus_pair_{pair_idx}"
                            pair_metadata = {
                                "query_file": str(pair_data["query_file"]),
                                "query_realpath": str(
                                    file_utils.realpath(pair_data["query_file"])
                                ),
                                "corpus_file": str(pair_data["corpus_file"]),
                                "corpus_realpath": str(
                                    file_utils.realpath(pair_data["corpus_file"])
                                ),
                            }
                            self.write_json(
                                obj=pair_metadata,
                                path=dst_dir.joinpath("file_pair_metadata.json"),
                            )
                        if self.args.output_qrel_format == "tsv":
                            res_path = dst_dir.joinpath("mined_qrels.tsv")
                        elif self.args.output_qrel_format == "grouped":
                            res_path = dst_dir.joinpath("mined_qrels.jsonl")
                        else:
                            raise ValueError

                        with file_utils.atomic_write(
                            file=res_path, root="parent"
                        ) as tfile:
                            file_utils.write_qrels(
                                qrels=pair_data["qrel"],
                                filepath=tfile,
                                format=self.args.output_qrel_format,
                            )

            if not isinstance(eval_dataset, dict):
                # If original dataset was not a dict, you should not return the results as dict
                results = list(results.values())[0]
        else:
            results = None

        self.barrier()
        if broadcast_output:
            # copy the output from process 0 to all other processes
            logger.info("Broadcast output to all processes.")
            results = self.broadcast_obj(obj=results, src=0)
        return results

    def remove_temp_cache_pardir(self) -> None:
        """Delete cache pardir and all its content if it was supposed to be temporary.

        Does not raise exception on failure.
        """
        if not self.args.cleanup_temp_artifacts:
            return

        # Make sure all processes are done before attempting to delete the temp dir.
        self.barrier()
        if self.is_main_process:
            if self.cache_pardir is None:
                # There is nothing to remove
                return
            if not Path(self.cache_pardir).exists():
                # There is nothing to remove
                return
            if self.args.encoding_cache_dir is not None and Path(
                self.args.encoding_cache_dir
            ).samefile(Path(self.cache_pardir)):
                # It is not a temporary cache created by evaluator
                return

            # It is a temp dir and should be removed
            logger.debug(
                f"Attempting to delete the temporary cache dir: {rpath(self.cache_pardir)}"
            )
            file_utils.safe_remove_path(self.cache_pardir)



from . import encoder_contriever, encoder_default, losses
from .loss_base import RetrievalLoss
from .model_args import ModelArguments
from .pretrained_encoder import PretrainedEncoder
from .pretrained_retriever import PretrainedRetriever, RetrieverOutput
from .retriever_biencoder import BiEncoderRetriever



import copy
import re
from collections import defaultdict
from typing import Dict, List, Optional, Set, Union

import pytrec_eval


def get_measures_max_cutoff(measures: Union[str, List[str], Set[str]]) -> int:
    """Returns Maximum cutoff value among several metrics.

    Args:
        measures (Union[str, List[str], Set[str]]): a collection of measures (e.g., ``mrr_cut_10,ndcg_cut_10``).

    Returns:
        maximum cutoff value among these measures.
    """
    measures = expand_and_combine_eval_measures(
        measures=copy.deepcopy(measures),
        additional_supported_measures="mrr_cut",
        reconstruct=False,
    )
    max_k = -1
    for k_values in measures.values():
        for k in k_values:
            if int(k) == float(k):
                num_k = int(k)
            else:
                num_k = int(k) + 1
            if num_k > max_k:
                max_k = num_k
    return max_k


def expand_and_combine_eval_measures(
    measures: Union[str, List[str], Set[str]],
    supported_measures: Optional[Union[str, List[str], Set[str]]] = None,
    additional_supported_measures: Optional[Union[str, List[str], Set[str]]] = None,
    reconstruct: bool = True,
) -> Union[Set[str], Dict[str, Set[str]]]:
    """Preprocess and combine all IR metric names into a standard format.

    Adapted combination of these two functions:

        * https://github.com/cvangysel/pytrec_eval/blob/0d8571efc2dea5a792358410d5affbe0ad383602/py/__init__.py#L66
        * https://github.com/cvangysel/pytrec_eval/blob/0d8571efc2dea5a792358410d5affbe0ad383602/py/__init__.py#L76

    Args:
        measures (Union[str, List[str], Set[str]]): IR metric names to process.
        supported_measures (Optional[Union[str, List[str], Set[str]]]): supported measure names.
            If not provided, use supported measure names from ``pytrec_eval``. I.e., ``pytrec_eval.supported_measures``.
        additional_supported_measures (Optional[Union[str, List[str], Set[str]]]): Additional
            metric names to include in ``supported_measures``.
        reconstruct (bool): If True, reconstruct processed metric names into ``meas.p1,p2,p3`` format.
            If False, return a mapping from base metric names to a set of cuttoff values.

    Returns:
        Standardized metric names either formatted as strings or a mapping from base metric name to cutoff values.
    """
    if isinstance(measures, str):
        measures = {measures}
    elif isinstance(measures, list):
        measures = set(measures)
    elif not isinstance(measures, set):
        msg = f"'measures' should of type `Union[List[str], Set[str], str]`. Got: '{type(measures)}'"
        raise TypeError(msg)

    if supported_measures is None:
        supported_measures = copy.deepcopy(pytrec_eval.supported_measures)
    elif isinstance(supported_measures, str):
        supported_measures = {supported_measures}
    elif isinstance(supported_measures, list):
        supported_measures = set(supported_measures)
    elif not isinstance(supported_measures, set):
        msg = f"'supported_measures' should of type `Optional[Union[List[str], Set[str], str]]`. Got: '{type(supported_measures)}'"
        raise TypeError(msg)

    supported_nicknames = copy.deepcopy(pytrec_eval.supported_nicknames)

    if isinstance(additional_supported_measures, str):
        additional_supported_measures = {additional_supported_measures}

    if isinstance(additional_supported_measures, (list, set)):
        supported_measures = supported_measures.union(additional_supported_measures)
    elif additional_supported_measures is not None:
        msg = (
            f"'additional_supported_measures' should of type `Optional[Union[List[str], Set[str], str]]`."
            f" Got: '{type(additional_supported_measures)}'"
        )
        raise TypeError(msg)

    # from here: https://github.com/cvangysel/pytrec_eval/blob/0d8571efc2dea5a792358410d5affbe0ad383602/py/__init__.py#L66
    # Expand nicknames (e.g., official, all_trec)
    result = set()
    for _measure in measures:
        if _measure in supported_nicknames:
            result.update(supported_nicknames[_measure])
        else:
            result.add(_measure)
    measures = result

    # from here: https://github.com/cvangysel/pytrec_eval/blob/0d8571efc2dea5a792358410d5affbe0ad383602/py/__init__.py#L76
    RE_BASE = r"{}[\._]([0-9]+(\.[0-9]+)?(,[0-9]+(\.[0-9]+)?)*)"

    # break apart measures in any of the following formats and combine
    #  1) meas          -> {meas: {}}  # either non-parameterized measure or use default params
    #  2) meas.p1       -> {meas: {p1}}
    #  3) meas_p1       -> {meas: {p1}}
    #  4) meas.p1,p2,p3 -> {meas: {p1, p2, p3}}
    #  5) meas_p1,p2,p3 -> {meas: {p1, p2, p3}}
    param_meas = defaultdict(set)
    for measure in measures:
        if measure not in supported_measures and measure not in supported_nicknames:
            matches = (
                (m, re.match(RE_BASE.format(re.escape(m)), measure))
                for m in supported_measures
            )
            match = next(filter(lambda x: x[1] is not None, matches), None)
            if match is None:
                raise ValueError("unsupported measure {}".format(measure))
            base_meas, meas_args = match[0], match[1].group(1)
            param_meas[base_meas].update(meas_args.split(","))
        elif measure not in param_meas:
            param_meas[measure] = set()

    if not reconstruct:
        return param_meas

    # re-construct in meas.p1,p2,p3 format for trec_eval
    fmt_meas = set()
    for meas, meas_args in param_meas.items():
        if meas_args:
            meas = "{}.{}".format(meas, ",".join(sorted(meas_args)))
        fmt_meas.add(meas)

    return fmt_meas
